from gym.envs.registration import EnvRegistry

registry = EnvRegistry()

def register(id, **kwargs):
    return registry.register(id, **kwargs)

def make(id, **kwargs):
    return registry.make(id, **kwargs)

def spec(id):
    return registry.spec(id)


'''
Multi-period inventory management
Hector Perez, Christian Hubbs, Owais Sarwar
4/14/2020
'''

import gym
import itertools
import numpy as np
from scipy.stats import *
from or_gym.utils import assign_env_config
from collections import deque

class InvManagementMasterEnv(gym.Env):
    '''
    The supply chain environment is structured as follows:
    
    It is a multi-period multi-echelon production-inventory system for a single non-perishable product that is sold only
    in discrete quantities. Each stage in the supply chain consists of an inventory holding area and a production area.
    The exception are the first stage (retailer: only inventory area) and the last stage (raw material transformation
    plant: only production area, with unlimited raw material availability). The inventory holding area holds the inventory
    necessary to produce the material at that stage. One unit of inventory produces one unit of product at each stage.
    There are lead times between the transfer of material from one stage to the next. The outgoing material from stage i 
    is the feed material for production at stage i-1. Stages are numbered in ascending order: Stages = {0, 1, ..., M} 
    (i.e. m = 0 is the retailer). Production at each stage is bounded by the stage's production capacity and the available
    inventory.
        
    At the beginning of each time period, the following sequence of events occurs:
    
    0) Stages 0 through M-1 place replenishment orders to their respective suppliers. Replenishment orders are filled
        according to available production capacity and available inventory at the respective suppliers.
    1) Stages 0 through M-1 receive incoming inventory replenishment shipments that have made it down the product pipeline
        after the stage's respective lead time.
    2) Customer demand occurs at stage 0 (retailer). It is sampled from a specified discrete probability distribution.
    3) Demand is filled according to available inventory at stage 0.
    4) Option: one of the following occurs,
        a) Unfulfilled sales and replenishment orders are backlogged at a penalty. 
            Note: Backlogged sales take priority in the following period.
        b) Unfulfilled sales and replenishment orders are lost with a goodwill loss penalty. 
    5) Surpluss inventory is held at each stage at a holding cost.
        
    '''
    def __init__(self, *args, **kwargs):
        '''
        periods = [positive integer] number of periods in simulation.
        I0 = [non-negative integer; dimension |Stages|-1] initial inventories for each stage.
        p = [positive float] unit price for final product.
        r = [non-negative float; dimension |Stages|] unit cost for replenishment orders at each stage.
        k = [non-negative float; dimension |Stages|] backlog cost or goodwill loss (per unit) for unfulfilled orders (demand or replenishment orders).
        h = [non-negative float; dimension |Stages|-1] unit holding cost for excess on-hand inventory at each stage.
            (Note: does not include pipeline inventory).
        c = [positive integer; dimension |Stages|-1] production capacities for each suppliers (stages 1 through |Stage|).
        L = [non-negative integer; dimension |Stages|-1] lead times in betwen stages.
        backlog = [boolean] are unfulfilled orders backlogged? True = backlogged, False = lost sales.
        dist = [integer] value between 1 and 4. Specifies distribution for customer demand.
            1: poisson distribution
            2: binomial distribution
            3: uniform random integer
            4: geometric distribution
            5: user supplied demand values
        dist_param = [dictionary] named values for parameters fed to statistical distribution.
            poisson: {'mu': <mean value>}
            binom: {'n': <mean value>, 'p': <probability between 0 and 1 of getting the mean value>}
            raindint: {'low' = <lower bound>, 'high': <upper bound>}
            geom: {'p': <probability. Outcome is the number of trials to success>}
        alpha = [float in range (0,1]] discount factor to account for the time value of money
        seed_int = [integer] seed for random state.
        user_D = [list] user specified demand for each time period in simulation
        '''
        # set default (arbitrary) values when creating environment (if no args or kwargs are given)
        self.periods = 30
        self.I0 = [100, 100, 200]
        self.p = 2
        self.r = [1.5, 1.0, 0.75, 0.5]
        self.k = [0.10, 0.075, 0.05, 0.025]
        self.h = [0.15, 0.10, 0.05]
        self.c = [100, 90, 80]
        self.L = [3, 5, 10]
        self.backlog = True
        self.dist = 1
        self.dist_param = {'mu': 20}
        self.alpha = 0.97
        self.seed_int = 0
        self.user_D = np.zeros(self.periods)
        self._max_rewards = 2000
        
        # add environment configuration dictionary and keyword arguments
        assign_env_config(self, kwargs)
        
        # input parameters
        try:
            self.init_inv = np.array(list(self.I0))
        except:
            self.init_inv = np.array([self.I0])
        self.num_periods = self.periods
        self.unit_price = np.append(self.p,self.r[:-1]) # cost to stage 1 is price to stage 2
        self.unit_cost = np.array(self.r)
        self.demand_cost = np.array(self.k)
        self.holding_cost = np.append(self.h,0) # holding cost at last stage is 0
        try:
            self.supply_capacity = np.array(list(self.c))
        except:
            self.supply_capacity = np.array([self.c])
        try:
            self.lead_time = np.array(list(self.L))
        except:
            self.lead_time = np.array([self.L])
        self.discount = self.alpha
        self.user_D = np.array(list(self.user_D))
        self.num_stages = len(self.init_inv) + 1
        m = self.num_stages
        lt_max = self.lead_time.max()
        
        #  parameters
        #  dictionary with options for demand distributions
        distributions = {1:poisson,
                         2:binom,
                         3:randint,
                         4:geom,
                         5:self.user_D}

        # check inputs
        assert np.all(self.init_inv) >=0, "The initial inventory cannot be negative"
        try:
            assert self.num_periods > 0, "The number of periods must be positive. Num Periods = {}".format(self.num_periods)
        except TypeError:
            print('\n{}\n'.format(self.num_periods))
        assert np.all(self.unit_price >= 0), "The sales prices cannot be negative."
        assert np.all(self.unit_cost >= 0), "The procurement costs cannot be negative."
        assert np.all(self.demand_cost >= 0), "The unfulfilled demand costs cannot be negative."
        assert np.all(self.holding_cost >= 0), "The inventory holding costs cannot be negative."
        assert np.all(self.supply_capacity > 0), "The supply capacities must be positive."
        assert np.all(self.lead_time >= 0), "The lead times cannot be negative."
        assert (self.backlog == False) | (self.backlog == True), "The backlog parameter must be a boolean."
        assert m >= 2, "The minimum number of stages is 2. Please try again"
        assert len(self.unit_cost) == m, "The length of r is not equal to the number of stages."
        assert len(self.demand_cost) == m, "The length of k is not equal to the number of stages."
        assert len(self.holding_cost) == m, "The length of h is not equal to the number of stages - 1."
        assert len(self.supply_capacity) == m-1, "The length of c is not equal to the number of stages - 1."
        assert len(self.lead_time) == m-1, "The length of L is not equal to the number of stages - 1."
        assert self.dist in [1,2,3,4,5], "dist must be one of 1, 2, 3, 4, 5."
        if self.dist < 5:
            assert distributions[self.dist].cdf(0,**self.dist_param), "Wrong parameters given for distribution."
        else:
            assert len(self.user_D) == self.num_periods, "The length of the user specified distribution is not equal to the number of periods."
        assert (self.alpha>0) & (self.alpha<=1), "alpha must be in the range (0,1]."
        
        # select distribution
        self.demand_dist = distributions[self.dist]  
        
        # set random generation seed (unless using user demands)
        if self.dist < 5:
            self.seed(self.seed_int)

        # intialize
        self.reset()
        
        # action space (reorder quantities for each stage; list)
        # An action is defined for every stage (except last one)
        # self.action_space = gym.spaces.Tuple(tuple(
            # [gym.spaces.Box(0, i, shape=(1,)) for i in self.supply_capacity]))
        self.pipeline_length = (m-1)*(lt_max+1)
        self.action_space = gym.spaces.Box(
            low=np.zeros(m-1), high=self.supply_capacity, dtype=np.int16)
        # observation space (Inventory position at each echelon, which is any integer value)
        self.observation_space = gym.spaces.Box(
            low=-np.ones(self.pipeline_length)*self.supply_capacity.max()*self.num_periods*10,
            high=np.ones(self.pipeline_length)*self.supply_capacity.max()*self.num_periods, dtype=np.int32)

        # self.observation_space = gym.spaces.Box(
        #     low=-np.ones(m-1)*self.supply_capacity.max()*self.num_periods*10, 
        #     high=self.supply_capacity*self.num_periods, dtype=np.int32)

    def seed(self,seed=None):
        '''
        Set random number generation seed
        '''
        # seed random state
        if seed != None:
            np.random.seed(seed=int(seed))
        
    def _RESET(self):
        '''
        Create and initialize all variables and containers.
        Nomenclature:
            I = On hand inventory at the start of each period at each stage (except last one).
            T = Pipeline inventory at the start of each period at each stage (except last one).
            R = Replenishment order placed at each period at each stage (except last one).
            D = Customer demand at each period (at the retailer)
            S = Sales performed at each period at each stage.
            B = Backlog at each period at each stage.
            LS = Lost sales at each period at each stage.
            P = Total profit at each stage.
        '''
        periods = self.num_periods
        m = self.num_stages
        I0 = self.init_inv
        
        # simulation result lists
        self.I=np.zeros([periods + 1, m - 1]) # inventory at the beginning of each period (last stage not included since iventory is infinite)
        self.T=np.zeros([periods + 1, m - 1]) # pipeline inventory at the beginning of each period (no pipeline inventory for last stage)
        self.R=np.zeros([periods, m - 1]) # replenishment order (last stage places no replenishment orders)
        self.D=np.zeros(periods) # demand at retailer
        self.S=np.zeros([periods, m]) # units sold
        self.B=np.zeros([periods, m]) # backlog (includes top most production site in supply chain)
        self.LS=np.zeros([periods, m]) # lost sales
        self.P=np.zeros(periods) # profit
        
        # initializetion
        self.period = 0 # initialize time
        self.I[0,:]=np.array(I0) # initial inventory
        self.T[0,:]=np.zeros(m-1) # initial pipeline inventory 
        self.action_log = np.zeros((periods, m-1))
        # set state
        self._update_state()
        
        return self.state

    def _update_state(self):
        m = self.num_stages - 1
        t = self.period
        lt_max = self.lead_time.max()
        state = np.zeros(m*(lt_max + 1))
        # state = np.zeros(m)
        if t == 0:
            state[:m] = self.I0
        else:
            state[:m] = self.I[t]

        if t == 0:
            pass
        elif t >= lt_max:
            state[-m*lt_max:] += self.action_log[t-lt_max:t].flatten()
        else:
            state[-m*(t):] += self.action_log[:t].flatten()

        self.state = state.copy()
    
    def _update_base_stock_policy_state(self):
        '''
        Get current state of the system: Inventory position at each echelon
        Inventory at hand + Pipeline inventory - backlog up to the current stage 
        (excludes last stage since no inventory there, nor replenishment orders placed there).
        '''
        n = self.period
        m = self.num_stages
        if n>=1:
            IP = np.cumsum(self.I[n,:] + self.T[n,:] - self.B[n-1,:-1])
        else:
            IP = np.cumsum(self.I[n,:] + self.T[n,:])
        self.state = IP
    
    def _STEP(self,action):
        '''
        Take a step in time in the multiperiod inventory management problem.
        action = [integer; dimension |Stages|-1] number of units to request from suppliers (last stage makes no requests)
        '''
        R = np.maximum(action, 0).astype(int)

        # get inventory at hand and pipeline inventory at beginning of the period
        n = self.period
        L = self.lead_time
        I = self.I[n,:].copy() # inventory at start of period n
        T = self.T[n,:].copy() # pipeline inventory at start of period n
        m = self.num_stages # number of stages
        
        # get production capacities
        c = self.supply_capacity # capacity
        self.action_log[n] = R.copy()
        # available inventory at the m+1 stage (note: last stage has unlimited supply)
        Im1 = np.append(I[1:], np.Inf) 
        
        # place replenishment order
        if n>=1: # add backlogged replenishment orders to current request
            R = R + self.B[n-1,1:]
        Rcopy = R.copy() # copy original replenishment quantity
        R[R>=c] = c[R>=c] # enforce capacity constraint
        R[R>=Im1] = Im1[R>=Im1] # enforce available inventory constraint
        self.R[n,:] = R # store R[n]
        
        # receive inventory replenishment placed L periods ago
        RnL = np.zeros(m-1) # initialize
        for i in range(m-1):
            if n - L[i] >= 0:
                RnL[i] = self.R[n-L[i],i].copy() # replenishment placed at the end of period n-L-1
                I[i] = I[i] + RnL[i]
            
        # demand is realized
        if self.dist < 5:
            D0 = self.demand_dist.rvs(**self.dist_param)
        else:
            D0 = self.demand_dist[n] # user specified demand
        D = D0 # demand
        self.D[n] = D0 # store D[n]
        
        # add previous backlog to demand
        if n >= 1:
            D = D0 + self.B[n-1,0].copy() # add backlogs to demand
        
        # units sold
        S0 = min(I[0],D) # at retailer
        S = np.append(S0,R) # at each stage
        self.S[n,:] = S # store S[n]
        
        # update inventory on hand and pipeline inventory
        I = I - S[:-1] # updated inventory at all stages (exclude last stage)
        T = T - RnL + R # updated pipeline inventory at all stages (exclude last one)
        self.I[n+1,:] = I # store inventory available at start of period n + 1 (exclude last stage)
        self.T[n+1,:] = T # store pipeline inventory at start of period n + 1
        
        # unfulfilled orders
        U = np.append(D, Rcopy) - S # unfulfilled demand and replenishment orders
        
        # backlog and lost sales
        if self.backlog:
            B = U
            LS = np.zeros(m)
        else:
            LS = U # lost sales
            B = np.zeros(m)
        self.B[n,:] = B # store B[n]
        self.LS[n,:] = LS # store LS[n]

        # calculate profit
        p = self.unit_price 
        r = self.unit_cost 
        k = self.demand_cost
        h = self.holding_cost
        a = self.discount
        II = np.append(I,0) # augment inventory so that last has no onsite inventory
        RR = np.append(R,S[-1]) # augment replenishment orders to include production cost at last stage
        P = a**n*np.sum(p*S - (r*RR + k*U + h*II)) # discounted profit in period n
        # P = a**n*np.sum(p*S - (r*RR + k*U + h*I))
        self.P[n] = P # store P
        
        # update period
        self.period += 1  
        
        # update stae
        self._update_state()
        
        # set reward (profit from current timestep)
        reward = P 
        
        # determine if simulation should terminate
        if self.period >= self.num_periods:
            done = True
        else:
            done = False
            
        return self.state, reward, done, {}
    
    def sample_action(self):
        '''
        Generate an action by sampling from the action_space
        '''
        return self.action_space.sample()
        
    def base_stock_action(self,z):
        '''
        Sample action (number of units to request) based on a base-stock policy (order up to z policy)
        z = [integer list; dimension |Stages| - 1] base stock level (no inventory at the last stage)
        '''
        n = self.period
        c = self.supply_capacity
        m = self.num_stages
        IP = self._update_base_stock_policy_state() # extract inventory position (current state)
        
        try:
            dimz = len(z)
        except:
            dimz = 1
        assert dimz == m-1, "Wrong dimension on base stock level vector. Should be # Stages - 1."
        
        # calculate total inventory position at the beginning of period n
        R = z - IP # replenishmet order to reach zopt

        # check if R can actually be fulfilled (capacity and inventory constraints)
        Im1 = np.append(self.I[n,1:], np.Inf) # available inventory at the m+1 stage
                                            # NOTE: last stage has unlimited raw materials
        Rpos = np.column_stack((np.zeros(len(R)),R)) # augmented materix to get replenishment only if positive
        A = np.column_stack((c, np.max(Rpos,axis=1), Im1)) # augmented matrix with c, R, and I_m+1 as columns
        
        R = np.min(A, axis = 1) # replenishmet order to reach zopt (capacity constrained)
        
        return R

    def step(self, action):
        return self._STEP(action)

    def reset(self):
        return self._RESET()
        
class InvManagementBacklogEnv(InvManagementMasterEnv):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
class InvManagementLostSalesEnv(InvManagementMasterEnv):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.backlog = False
        self.observation_space = gym.spaces.Box(
            low=np.zeros(self.pipeline_length), # Never goes negative without backlog
            high=np.ones(self.pipeline_length)*self.supply_capacity.max()*self.num_periods, dtype=np.int32)


'''
Example taken from Balaji et al.
Paper: https://arxiv.org/abs/1911.10641
GitHub: https://github.com/awslabs/or-rl-benchmarks
'''
import gym
from gym import spaces
import itertools
import numpy as np
from collections.abc import Iterable
from or_gym.utils import assign_env_config

class NewsvendorEnv(gym.Env):
    '''
    Multi-Period Newsvendor with Lead Times

    The MPNV requires meeting stochastic demand by having sufficient
    inventory on hand to satisfy customers. The inventory orders are not
    instantaneous and have multi-period leadtimes. Additionally, there are
    costs associated with holding unsold inventory, however unsold inventory
    expires at the end of each period.

    Observation:
        Type: Box 
        State Vector: S = (p, c, h, k, mu, x_l, x_l-1)
        p = price
        c = cost
        h = holding cost
        k = lost sales penalty
        mu = mean of demand distribution
        x_l = order quantities in the queue

    Actions:
        Type: Box
        Amount of product to order.

    Reward:
        Sales minus discounted purchase price, minus holding costs for
        unsold product or penalties associated with insufficient inventory.

    Initial State:
        Parameters p, c, h, k, and mu, with no inventory in the pipeline.

    Episode Termination:
        Termination occurs after the maximum number of time steps is reached
        (40 by default).
    '''
    def __init__(self, *args, **kwargs):
        self.lead_time = 5
        self.max_inventory = 4000
        self.max_order_quantity = 2000
        self.step_limit = 40
        self.p_max = 100    # Max sale price
        self.h_max = 5      # Max holding cost
        self.k_max = 10     # Max lost sales penalty
        self.mu_max = 200   # Max mean of the demand distribution
        self.gamma = 1      # Discount factor
        assign_env_config(self, kwargs)

        self.obs_dim = self.lead_time + 5
        self.observation_space = spaces.Box(
            low=np.zeros(self.obs_dim),
            high=np.array(
                [self.p_max, self.p_max, self.h_max, self.k_max, self.mu_max] +
                [self.max_order_quantity] * self.lead_time),
            dtype=np.float32)
        self.action_space = spaces.Box(
            low=np.zeros(1), high=np.array([self.max_order_quantity]), 
            dtype=np.float32)

        self.reset()

    def _STEP(self, action):
        done = False
        order_qty = max(0, # Ensure order > 0
            min(action, self.max_inventory - self.state[5:].sum())) # Cap inventory
        demand = np.random.poisson(self.mu)
        inventory = self.state[5:]
        if self.lead_time == 0: # No lead time -> instant fulfillment
            inv_on_hand = order_qty
        else:
            inv_on_hand = inventory[0]
        sales = min(inv_on_hand, demand) * self.price
        excess_inventory = max(0, inv_on_hand - demand)
        short_inventory = max(0, demand - inv_on_hand)
        purchase_cost = excess_inventory * self.cost * order_qty * \
            self.gamma ** self.lead_time
        holding_cost = excess_inventory * self.h
        lost_sales_penalty = short_inventory * self.k
        reward = sales - purchase_cost - holding_cost - lost_sales_penalty

        # Update state, note inventory on hand expires at each time step
        new_inventory = np.zeros(self.lead_time)
        new_inventory[:-1] += inventory[1:]
        new_inventory[-1] += order_qty
        self.state = np.hstack([self.state[:5], new_inventory])

        self.step_count += 1
        if self.step_count >= self.step_limit:
            done = True
        if isinstance(reward, Iterable):
            # TODO: Sometimes reward is np.array with one entry
            reward = sum(reward)

        return self.state, reward, done, {}

    def _RESET(self):
        # Randomize costs
        self.price = max(1, np.random.rand() * self.p_max)
        self.cost = max(1, np.random.rand() * self.price)
        self.h = np.random.rand() * min(self.cost, self.h_max)
        self.k = np.random.rand() * self.k_max
        self.mu = np.random.rand() * self.mu_max
        self.state = np.zeros(self.obs_dim)
        self.state[:5] = np.array([self.price, self.cost, self.h,
            self.k, self.mu])

        self.step_count = 0

        return self.state

    def reset(self):
        return self._RESET()

    def step(self, action):
        return self._STEP(action)


import numpy as np
import gym
from gym import spaces, logger
from gym.utils import seeding
from or_gym.utils import assign_env_config
import copy

class KnapsackEnv(gym.Env):
    '''
    Unbounded Knapsack Problem

    The Knapsack Problem (KP) is a combinatorial optimization problem which
    requires the user to select from a range of goods of different values and
    weights in order to maximize the value of the selected items within a 
    given weight limit. This version is unbounded meaning that we can select
    items without limit. 

    The episodes proceed by selecting items and placing them into the
    knapsack one at a time until the weight limit is reached or exceeded, at
    which point the episode ends.

    Observation:
        Type: Tuple, Discrete
        0: list of item weights
        1: list of item values
        2: maximum weight of the knapsack
        3: current weight in knapsack

    Actions:
        Type: Discrete
        0: Place item 0 into knapsack
        1: Place item 1 into knapsack
        2: ...

    Reward:
        Value of item successfully placed into knapsack or 0 if the item
        doesn't fit, at which point the episode ends.

    Starting State:
        Lists of available items and empty knapsack.

    Episode Termination:
        Full knapsack or selection that puts the knapsack over the limit.
    '''
    def __init__(self, *args, **kwargs):
        # Generate data with consistent random seed to ensure reproducibility
        self.N = 200
        self.max_weight = 200
        self.current_weight = 0
        self._max_reward = 10000
        self.mask = True
        self.seed = 0
        self.item_numbers = np.arange(self.N)
        self.item_weights = np.random.randint(1, 100, size=self.N)
        self.item_values = np.random.randint(0, 100, size=self.N)
        self.over_packed_penalty = 0
        self.randomize_params_on_reset = False
        # Add env_config, if any
        assign_env_config(self, kwargs)
        self.set_seed()
        
        obs_space = spaces.Box(
            0, self.max_weight, shape=(2*self.N + 1,), dtype=np.int16)
        self.action_space = spaces.Discrete(self.N)
        if self.mask:
            self.observation_space = spaces.Dict({
                "action_mask": spaces.Box(0, 1, shape=(self.N,)),
                "avail_actions": spaces.Box(0, 1, shape=(self.N,)),
                "state": obs_space
                })
        else:
            self.observation_space = spaces.Box(
                0, self.max_weight, shape=(2, self.N + 1), dtype=np.int16)
        
        self.reset()
        
    def _STEP(self, item):
        # Check that item will fit
        if self.item_weights[item] + self.current_weight <= self.max_weight:
            self.current_weight += self.item_weights[item]
            reward = self.item_values[item]
            if self.current_weight == self.max_weight:
                done = True
            else:
                done = False
        else:
            # End trial if over weight
            reward = self.over_packed_penalty
            done = True
            
        self._update_state()
        return self.state, reward, done, {}
    
    def _get_obs(self):
        return self.state
    
    def _update_state(self):
        if self.mask:
            mask = np.where(self.current_weight + self.item_weights > self.max_weight,
            0, 1)
            state = np.hstack([
                self.item_weights,
                self.item_values,
                np.array([self.current_weight])
                ])
            self.state = {
                "action_mask": mask,
                "avail_actions": np.ones(self.N),
                "state": state
                }
        else:
            state = np.vstack([
                self.item_weights,
                self.item_values])
            self.state = np.hstack([
                state,
                np.array([
                    [self.max_weight],
                     [self.current_weight]])
                ])        
    
    def _RESET(self):
        if self.randomize_params_on_reset:
            self.item_weights = np.random.randint(1, 100, size=self.N)
            self.item_values = np.random.randint(0, 100, size=self.N)
        self.current_weight = 0
        self._update_state()
        return self.state
    
    def sample_action(self):
        return np.random.choice(self.item_numbers)

    def set_seed(self, seed=None):
        if seed == None:
            seed = np.random.randint(0, np.iinfo(np.int32).max)        
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def reset(self):
        return self._RESET()

    def step(self, action):
        return self._STEP(action)

class BinaryKnapsackEnv(KnapsackEnv):
    '''
    Binary Knapsack Problem

    The Binary or 0-1 KP allows selection of each item only once or not at
    all.

    The episodes proceed by selecting items and placing them into the
    knapsack one at a time until the weight limit is reached or exceeded, at
    which point the episode ends.

    Observation:
        Type: Tuple, Discrete
        0: list of item weights
        1: list of item values
        2: list of item limits
        3: maximum weight of the knapsack
        4: current weight in knapsack

    Actions:
        Type: Discrete
        0: Place item 0 into knapsack
        1: Place item 1 into knapsack
        2: ...

    Reward:
        Value of item successfully placed into knapsack or 0 if the item
        doesn't fit, at which point the episode ends.

    Starting State:
        Lists of available items and empty knapsack.

    Episode Termination:
        Full knapsack or selection that puts the knapsack over the limit.
    '''
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.item_weights = np.random.randint(1, 100, size=self.N)
        self.item_values = np.random.randint(0, 100, size=self.N)
        assign_env_config(self, kwargs)

        obs_space = spaces.Box(
            0, self.max_weight, shape=(3, self.N + 1), dtype=np.int32)
        if self.mask:
            self.observation_space = spaces.Dict({
                "action_mask": spaces.Box(0, 1, shape=(len(self.item_limits),)),
                "avail_actions": spaces.Box(0, 1, shape=(len(self.item_limits),)),
                "state": obs_space
            })
        else:
            self.observation_space = obs_space

        self.reset()

    def _STEP(self, item):
        # Check item limit
        if self.item_limits[item] > 0:
            # Check that item will fit
            if self.item_weights[item] + self.current_weight <= self.max_weight:
                self.current_weight += self.item_weights[item]
                reward = self.item_values[item]
                if self.current_weight == self.max_weight:
                    done = True
                else:
                    done = False
                self._update_state(item)
            else:
                # End if over weight
                reward = 0
                done = True
        else:
            # End if item is unavailable
            reward = 0
            done = True
            
        return self.state, reward, done, {}

    def _update_state(self, item=None):
        if item is not None:
            self.item_limits[item] -= 1
        state_items = np.vstack([
            self.item_weights,
            self.item_values,
            self.item_limits
        ])
        state = np.hstack([
            state_items, 
            np.array([[self.max_weight],
                      [self.current_weight], 
                      [0] # Serves as place holder
                ])
        ])
        if self.mask:
            mask = np.where(self.current_weight + self.item_weights > self.max_weight,
                0, 1)
            mask = np.where(self.item_limits > 0, mask, 0)
            self.state = {
                "action_mask": mask,
                "avail_actions": np.ones(self.N),
                "state": state
            }
        else:
            self.state = state.copy()
        
    def sample_action(self):
        return np.random.choice(
            self.item_numbers[np.where(self.item_limits!=0)])
    
    def _RESET(self):
        self.current_weight = 0
        self.item_limits = np.ones(self.N)
        self._update_state()
        return self.state

class BoundedKnapsackEnv(KnapsackEnv):
    '''
    Bounded Knapsack Problem

    The Knapsack Problem (KP) is a combinatorial optimization problem which
    requires the user to select from a range of goods of different values and
    weights in order to maximize the value of the selected items within a 
    given weight limit. This version is bounded meaning each item can be
    selected a limited number of times.

    The episodes proceed by selecting items and placing them into the
    knapsack one at a time until the weight limit is reached or exceeded, at
    which point the episode ends.

    Observation:
        Type: Tuple, Discrete
        0: list of item weights
        1: list of item values
        2: list of item limits
        3: maximum weight of the knapsack
        4: current weight in knapsack

    Actions:
        Type: Discrete
        0: Place item 0 into knapsack
        1: Place item 1 into knapsack
        2: ...

    Reward:
        Value of item successfully placed into knapsack or 0 if the item
        doesn't fit, at which point the episode ends.

    Starting State:
        Lists of available items and empty knapsack.

    Episode Termination:
        Full knapsack or selection that puts the knapsack over the limit.
    '''
    def __init__(self, *args, **kwargs):
        self.N = 200
        self.item_limits_init = np.random.randint(1, 10, size=self.N)
        self.item_limits = self.item_limits_init.copy()
        super().__init__()
        self.item_weights = np.random.randint(1, 100, size=self.N)
        self.item_values = np.random.randint(0, 100, size=self.N)

        assign_env_config(self, kwargs)
        
        obs_space = spaces.Box(
            0, self.max_weight, shape=(3, self.N + 1), dtype=np.int32)
        if self.mask:
            self.observation_space = spaces.Dict({
                "action_mask": spaces.Box(0, 1, shape=(len(self.item_limits),)),
                "avail_actions": spaces.Box(0, 1, shape=(len(self.item_limits),)),
                "state": obs_space
            })
        else:
            self.observation_space = obs_space
        
    def _STEP(self, item):
        # Check item limit
        if self.item_limits[item] > 0:
            # Check that item will fit
            if self.item_weights[item] + self.current_weight <= self.max_weight:
                self.current_weight += self.item_weights[item]
                reward = self.item_values[item]
                if self.current_weight == self.max_weight:
                    done = True
                else:
                    done = False
                self._update_state(item)
            else:
                # End if over weight
                reward = 0
                done = True
        else:
            # End if item is unavailable
            reward = 0
            done = True
            
        return self.state, reward, done, {}

    def _update_state(self, item=None):
        if item is not None:
            self.item_limits[item] -= 1
        state_items = np.vstack([
            self.item_weights,
            self.item_values,
            self.item_limits
        ])
        state = np.hstack([
            state_items, 
            np.array([[self.max_weight],
                      [self.current_weight], 
                      [0] # Serves as place holder
                ])
        ])
        if self.mask:
            mask = np.where(self.current_weight + self.item_weights > self.max_weight,
                0, 1)
            mask = np.where(self.item_limits > 0, mask, 0)
            self.state = {
                "action_mask": mask,
                "avail_actions": np.ones(self.N),
                "state": state
            }
        else:
            self.state = state.copy()
        
    def sample_action(self):
        return np.random.choice(
            self.item_numbers[np.where(self.item_limits!=0)])
    
    def _RESET(self):
        self.current_weight = 0
        self.item_limits = self.item_limits_init.copy()
        self._update_state()
        return self.state

class OnlineKnapsackEnv(BoundedKnapsackEnv):
    '''
    Online Knapsack Problem

    The Knapsack Problem (KP) is a combinatorial optimization problem which
    requires the user to select from a range of goods of different values and
    weights in order to maximize the value of the selected items within a 
    given weight limit. This version is online meaning each item is randonly
    presented to the algorithm one at a time, at which point the algorithm 
    can either accept or reject the item. After seeing a fixed number of 
    items are shown, the episode terminates. If the weight limit is reached
    before the episode ends, then it terminates early.

    Observation:
        Type: Tuple, Discrete
        0: list of item weights
        1: list of item values
        2: list of item limits
        3: maximum weight of the knapsack
        4: current weight in knapsack


    Actions:
        Type: Discrete
        0: Reject item
        1: Place item into knapsack

    Reward:
        Value of item successfully placed into knapsack or 0 if the item
        doesn't fit, at which point the episode ends.

    Starting State:
        Lists of available items and empty knapsack.

    Episode Termination:
        Full knapsack, selection that puts the knapsack over the limit, or
        the number of items to be drawn has been reached.
    '''
    def __init__(self, *args, **kwargs):
        BoundedKnapsackEnv.__init__(self)
        self.item_weights = np.random.randint(1, 100, size=self.N)
        self.item_values = np.random.randint(0, 100, size=self.N)
        assign_env_config(self, kwargs)
        self.action_space = spaces.Discrete(2)

        obs_space = spaces.Box(0, self.max_weight, shape=(4,))
        if self.mask:
            self.observation_space = spaces.Dict({
                'state': obs_space,
                'avail_actions': spaces.Box(0, 1, shape=(2,)),
                'action_mask': spaces.Box(0, 1, shape=(2,))
            })
        else:
            self.observation_space = obs_space

        self.step_counter = 0
        self.step_limit = 50
        
        self.state = self.reset()
        self._max_reward = 600
        
    def _STEP(self, action):
        if bool(action):
            # Check that item will fit
            if self.item_weights[self.current_item] + self.current_weight <= self.max_weight:
                self.current_weight += self.item_weights[self.current_item]
                reward = self.item_values[self.current_item]
                if self.current_weight == self.max_weight:
                    done = True
                else:
                    done = False
                self._update_state()
            else:
                # End if over weight
                reward = 0
                done = True
        else:
            reward = 0
            done = False
            
        self.step_counter += 1
        if self.step_counter >= self.step_limit:
            done = True
            
        return self.state, reward, done, {}
    
    def _update_state(self):
        self.current_item = np.random.choice(self.item_numbers, p=self.item_probs)
        current_item_weight = self.item_weights[self.current_item]
        state = np.array([
                self.current_weight,
                self.current_item,
                current_item_weight,
                self.item_values[self.current_item]
                ])
        if self.mask:
            mask = np.ones(2)
            if current_item_weight + self.current_weight > self.max_weight:
                mask[1] = 0
            self.state = {
                'state': state,
                'avail_actions': np.ones(2),
                'action_mask': mask
            }            
        else:
            self.state = state

    def sample_action(self):
        return np.random.choice([0, 1])
    
    def _RESET(self):
        if not hasattr(self, 'item_probs'):
            self.item_probs = self.item_limits_init / self.item_limits_init.sum()
        self.current_weight = 0
        self.step_counter = 0
        self.item_limits = self.item_limits_init.copy()
        self._update_state()
        return self.state


import numpy as np
import gym
from gym import spaces
from or_gym import utils
from copy import copy, deepcopy
import matplotlib.pyplot as plt

class TSPEnv(gym.Env):
    '''
    Bi-directional connections and uniform cost

    This version of the TSP uses a sparse graph with uniform cost.
    The goal is to minimize the cost to traverse all of the nodes in the
    network. All connections are bi-directional meaning if a connection
    between nodes n and m exist, then the agent can move in either direction.
    The network is randomly generated with N nodes when the environment is
    initialized using or_gym.make(). 
    
    TSP-v0 allows repeat visits to nodes with no additional penalty beyond
    the nominal movement cost.

    Observation:
        

    Actions:
        Type: Discrete
        0: move to node 0
        1: move to node 1
        2: ...

    Action Masking (optional):
        Masks non-existent connections, otherwise a large penalty is imposed
        on the agent.

    Reward:
        Cost of moving from node to node or large negative penalty for
        attempting to move to a node via a non-existent connection.

    Starting State:
        Random node

    Episode Termination:
        All nodes have been visited or the maximimum number of steps (2N)
        have been reached.
    '''
    def __init__(self, *args, **kwargs):
        self.N = 50
        self.move_cost = -1
        self.invalid_action_cost = -100
        self.mask = False
        utils.assign_env_config(self, kwargs)

        self.nodes = np.arange(self.N)
        self.step_limit = 2*self.N
        self.obs_dim = 1+self.N**2
        obs_space = spaces.Box(-1, self.N, shape=(self.obs_dim,), dtype=np.int32)
        if self.mask:
            self.observation_space = spaces.Dict({
                "action_mask": spaces.Box(0, 1, shape=(self.N,), dtype=np.int8),
                "avail_actions": spaces.Box(0, 1, shape=(self.N,), dtype=np.int8),
                "state": obs_space
            })
        else:
            self.observation_space = obs_space
        self.action_space = spaces.Discrete(self.N)
        
        self.reset()
        
    def _STEP(self, action):
        done = False
        connections = self.node_dict[self.current_node]
        # Invalid action
        if action not in connections:
            reward = self.invalid_action_cost
        # Move to new node
        else:
            self.current_node = action
            reward = self.move_cost
            self.visit_log[self.current_node] += 1
            
        self.state = self._update_state()
        self.step_count += 1
        # See if all nodes have been visited
        unique_visits = sum([1 if v > 0 else 0 
            for v in self.visit_log.values()])
        if unique_visits >= self.N:
            done = True
            reward += 1000
        if self.step_count >= self.step_limit:
            done = True
            
        return self.state, reward, done, {}
        
    def _RESET(self):
        self.step_count = 0
        self._generate_connections()
        self.current_node = np.random.choice(self.nodes)
        self.visit_log = {n: 0 for n in self.nodes}
        self.visit_log[self.current_node] += 1
        
        self.state = self._update_state()
        return self.state
        
    def _update_state(self):
        node_connections = self.adjacency_matrix.copy()
        # Set value to 1 for existing, un-visited nodes
        # Set value to -1 for existing, visited nodes
        # Set value to 0 if connection doesn't exist
        visited = np.array([bool(min(v, 1))
            for v in self.visit_log.values()])
        node_connections[:, visited] = -1
        node_connections[np.where(self.adjacency_matrix==0)] = 0

        connections = node_connections.flatten().astype(int)
        obs = np.hstack([self.current_node, connections])
        if self.mask:
            mask = node_connections[self.current_node]
            # mask = np.array([1 if c==1 and v==0 else 0 
            #     for c, v in zip(cons_from_node, self.visit_log.values())])
            state = {
                "action_mask": mask,
                "avail_actions": np.ones(self.N),
                "state": obs,
            }
        else:
            state = obs.copy()

        return state
        
    def _generate_connections(self):
        node_dict = {}
        for n in range(self.N):
            connections = np.random.randint(2, self.N - 1)
            node_dict[n] = np.sort(
               np.random.choice(self.nodes[np.where(self.nodes!=n)],
                                 size=connections, replace=False))
        # Get unique, bi-directional connections
        for k, v in node_dict.items():
            for k1, v1 in node_dict.items():
                if k == k1:
                    continue
                if k in v1 and k1 not in v:
                    v = np.append(v, k1)

            node_dict[k] = np.sort(v.copy())
        self.node_dict = deepcopy(node_dict)
        self._generate_adjacency_matrix()
    
    def _generate_adjacency_matrix(self):
        self.adjacency_matrix = np.zeros((self.N, self.N))
        for k, v in self.node_dict.items():
            self.adjacency_matrix[k][v] += 1
        self.adjacency_matrix.astype(int)
            
    def _generate_coordinates(self):
        n = np.linspace(0, 2*np.pi, self.N+1)
        x = np.cos(n)
        y = np.sin(n)
        return np.vstack([x, y])

    def _get_node_distance(self, N0, N1):
        return np.sqrt(np.power(N0[0] - N1[0], 2) + np.power(N0[1] - N1[1], 2))
            
    def plot_network(self, offset=(0.02, 0.02)):
        coords = self._generate_coordinates()
        fig, ax = plt.subplots(figsize=(12,8))
        ax.scatter(coords[0], coords[1], s=40)
        for n, c in self.node_dict.items():
            for k in c:
                line = np.vstack([coords[:, n], coords[:, k]])
                dis = self._get_node_distance(line[0], line[1])
                # dis = np.sqrt(np.power(line[0, 0] - line[1, 0], 2) + 
                #               np.power(line[0, 1] - line[1, 1], 2))
                ax.plot(line[:,0], line[:,1], c='g', zorder=-1)
        #         ax.arrow(line[0, 0], line[0, 1], line[1, 0], line[1, 1])
            ax.annotate(r"$N_{:d}$".format(n), xy=(line[0]+offset), zorder=2)
        ax.xaxis.set_visible(False)
        ax.yaxis.set_visible(False)
        plt.show()

    def step(self, action):
        return self._STEP(action)

    def reset(self):
        return self._RESET()

class TSPDistCost(TSPEnv):
    '''
    Fully connected network with distance-based cost.

    This environment enables travel between all nodes in the network and 
    incurs cost based on the Euclidean distance between nodes. The goal is to
    minimize the cost to traverse all of the nodes in the network exactly 
    once. The agent incurs a large penalty and ends the episode if it moves to 
    a node more than once. All connections are bi-directional meaning if a 
    connection between nodes n and m exist, then the agent can move in either 
    direction. The network is randomly generated with N nodes when the 
    environment is initialized using or_gym.make(). 
    
    Observation:
        Type: Box
        0: Current Node
        1: 0 or 1 if node 0 has been visited or not
        2: 0 or 1 if node 1 has been visited or not
        3: ...

    Actions:
        Type: Discrete
        0: move to node 0
        1: move to node 1
        2: ...

    Action Masking (optional):
        Masks visited nodes.

    Reward:
        Cost of moving from node to node.

    Starting State:
        Random node

    Episode Termination:
        All nodes have been visited or a node has been visited again.
    '''
    def __init__(self, *args, **kwargs):
        self.N = 50
        self.invalid_action_cost = -100
        self.mask = False
        utils.assign_env_config(self, kwargs)
        self.nodes = np.arange(self.N)
        self.coords = self._generate_coordinates()
        self.distance_matrix = self._get_distance_matrix()

        self.obs_dim = 1+self.N
        obs_space = spaces.Box(-1, self.N, shape=(self.obs_dim,), dtype=np.int32)
        if self.mask:
            self.observation_space = spaces.Dict({
                "action_mask": spaces.Box(0, 1, shape=(self.N,), dtype=np.int8),
                "avail_actions": spaces.Box(0, 1, shape=(self.N,), dtype=np.int8),
                "state": obs_space
            })
        else:
            self.observation_space = obs_space

        self.action_space = spaces.Discrete(self.N)
        
        self.reset()

    def _STEP(self, action):
        done = False
        if self.visit_log[action] > 0:
            # Node already visited
            reward = self.invalid_action_cost
            done = True
        else:
            reward = self.distance_matrix[self.current_node, action]
            self.current_node = action
            self.visit_log[self.current_node] = 1
            
        self.state = self._update_state()
        # See if all nodes have been visited
        unique_visits = self.visit_log.sum()
        if unique_visits == self.N:
            done = True
            
        return self.state, reward, done, {}

    def _RESET(self):
        self.step_count = 0
        self.current_node = np.random.choice(self.nodes)
        self.visit_log = np.zeros(self.N)
        self.visit_log[self.current_node] += 1
        
        self.state = self._update_state()
        return self.state

    def _generate_coordinates(self):
        return np.vstack([np.random.rand(self.N), np.random.rand(self.N)])

    def _get_distance_matrix(self):
        # Distance matrix
        distance_matrix = np.zeros((self.N, self.N))
        for i in range(self.N):
            # Take advantage of symmetrical matrix
            for j in range(self.N):
                if j <= i:
                    continue
                d = self._get_node_distance(self.coords[:, i], self.coords[:, j])
                distance_matrix[i, j] += d
                
        distance_matrix += distance_matrix.T
        return distance_matrix

    def _update_state(self):
        mask = np.where(self.visit_log==0, 0 , 1)
        obs = np.hstack([self.current_node, mask])
        if self.mask:
            state = {
                "avail_actions": np.ones(self.N),
                "action_mask": mask,
                "state": obs
            }
        else:
            state = obs.copy()
        return state

    def step(self, action):
        return self._STEP(action)

    def reset(self):
        return self._RESET()
    


'''
Multi-period inventory management
Hector Perez, Christian Hubbs, Can Li
9/14/2020
'''

import gym
import itertools
import numpy as np
import networkx as nx
import pandas as pd
from scipy.stats import *
from or_gym.utils import assign_env_config
from collections import deque
import matplotlib.pyplot as plt

class NetInvMgmtMasterEnv(gym.Env):
    '''
    The supply network environment is structured as follows:
    
    It is a multi-period multi-node production-inventory system for 
    a single non-perishable product that is sold in discrete quantities. 
    Two main types of nodes exist: 1) production nodes (which have an 
    inventory holding area and a manufacturing area), and 2) distribution
    nodes (which only have an inventory holding area). Retail nodes are
    considered distribution nodes. Other node types in the network include 
    1) raw material nodes (source nodes), which have unlimited supply
    of raw materials, and 2) market nodes (sink nodes), which generate an 
    uncertain demand on their respective retailers in each period. 

    Within production nodes, the inventory holding area holds the inventory 
    necessary to produce the respective intermediate material at that node. 
    Yield ratios are specified at each production stage relating the amount 
    of material produced from one unit of inventory. Production at each node
    is bounded by the nodes's production capacity and the available inventory.

    Lead times between neighbor nodes exist and are associated with the edges
    connecting them.
        
    At the beginning of each time period, the following sequence of events occurs:
    
    0) Each node places replenishment orders to their immediate suppliers. 
       Replenishment orders are filled according to available production capacity 
       and available inventory at the suppliers. There is a cost associated with
       each order request.
    1) Incoming inventory replenishment shipments that have made it down network
       pipeline (after the associated lead time) are received at each node.
    2) Market demands occur at the retail nodes. Demands are sampled from a 
       specified discrete probability distribution. Demands are filled according 
       to available inventory at the retailers.
    4) Option: one of the following occurs,
        a) Unfulfilled sales are backlogged at a penalty. 
            Note: Backlogged orders are added to the next period's market demand.
        b) Unfulfilled sales and replenishment orders are lost with a 
           goodwill-loss penalty. 
    5) Surpluss inventory is held at each stage at a holding cost. 
        Pipeline inventories also incur a cost for each period in the pipeline.
        
    '''
    def __init__(self, *args, **kwargs):
        '''
        num_periods = number of periods in simulation.
        Node specific parameters:
            - I0 = initial inventory.
            - C = production capacity.
            - v = production yield in the range (0, 1].
            - o = unit operating cost (feed-based)
            - h = unit holding cost for excess on-hand inventory.
        Edge specific parameters:
            - L = lead times in betwen adjacent nodes.
            - p = unit price to send material between adjacent nodes (purchase price/reorder cost)
            - b = unit backlog cost or good-wil loss for unfulfilled market demand between adjacent retailer and market.
            - g = unit holding cost for pipeline inventory on a specified edge.
            - prob_dist = probability distribution function on a (retailer, market) edge.
            - demand_dist = demand distribution for (retailer, market) edge. Two options:
                - use scipy probability distribution: must be a lambda function calling the rvs method of the distribution
                    i.e. lambda: poisson.rvs(mu=20)
                - use a list of user specified demands for each period. 
        backlog = Are unfulfilled orders backlogged? True = backlogged, False = lost sales.
        demand_dist = distribution function for customer demand (e.g. poisson, binomial, uniform, geometric, etc.)
        dist_param = named values for parameters fed to statistical distribution.
            poisson: {'mu': <mean value>}
            binom: {'n': <mean value>, 
                    'p': <probability between 0 and 1 of getting the mean value>}
            raindint: {'low' = <lower bound>, 'high': <upper bound>}
            geom: {'p': <probability. Outcome is the number of trials to success>}
        alpha = discount factor in the range (0,1] that accounts for the time value of money
        seed_int = integer seed for random state.
        user_D = dictionary containing user specified demand (list) for each (retail, market) pair at
            each time period in the simulation. If all zeros, ignored; otherwise, demands will be taken from this list.
        sample_path = dictionary specifying if is user_D (for each (retail, market) pair) is sampled from demand_dist.
        '''
        # set default (arbitrary) values when creating environment (if no args or kwargs are given)
        self._max_rewards = 2000
        self.num_periods = 30
        self.backlog = True
        self.alpha = 1.00
        self.seed_int = 0
        self.user_D = {(1,0): np.zeros(self.num_periods)}
        self.sample_path = {(1,0): False}
        self._max_rewards = 2000

        # create graph
        self.graph = nx.DiGraph()
        # Market 
        self.graph.add_nodes_from([0])
        # Retailer
        self.graph.add_nodes_from([1], I0 = 100,
                                        h = 0.030)
        # Distributors
        self.graph.add_nodes_from([2], I0 = 110,
                                        h = 0.020)
        self.graph.add_nodes_from([3], I0 = 80,
                                        h = 0.015)
        # Manufacturers
        self.graph.add_nodes_from([4], I0 = 400,
                                        C = 90,
                                        o = 0.010,
                                        v = 1.000,
                                        h = 0.012)
        self.graph.add_nodes_from([5], I0 = 350,
                                        C = 90,
                                        o = 0.015,
                                        v = 1.000,
                                        h = 0.013)
        self.graph.add_nodes_from([6], I0 = 380,
                                        C = 80,
                                        o = 0.012,
                                        v = 1.000,
                                        h = 0.011)
        # Raw materials
        self.graph.add_nodes_from([7, 8])
        # Links
        self.graph.add_edges_from([(1,0,{'p': 2.000,
                                         'b': 0.100,
                                         'demand_dist': poisson,
                                         'dist_param': {'mu': 20}}),
                                   (2,1,{'L': 5,
                                         'p': 1.500,
                                         'g': 0.010}),
                                   (3,1,{'L': 3,
                                         'p': 1.600,
                                         'g': 0.015}),
                                   (4,2,{'L': 8,
                                         'p': 1.000,
                                         'g': 0.008}),
                                   (4,3,{'L': 10,
                                         'p': 0.800,
                                         'g': 0.006}),
                                   (5,2,{'L': 9,
                                         'p': 0.700,
                                         'g': 0.005}),
                                   (6,2,{'L': 11,
                                         'p': 0.750,
                                         'g': 0.007}),
                                   (6,3,{'L': 12,
                                         'p': 0.800,
                                         'g': 0.004}),
                                   (7,4,{'L': 0,
                                         'p': 0.150,
                                         'g': 0.000}),
                                   (7,5,{'L': 1,
                                         'p': 0.050,
                                         'g': 0.005}),
                                   (8,5,{'L': 2,
                                         'p': 0.070,
                                         'g': 0.002}),
                                   (8,6,{'L': 0,
                                         'p': 0.200,
                                         'g': 0.000})])
        
        # add environment configuration dictionary and keyword arguments
        assign_env_config(self, kwargs)

        # Save user_D and sample_path to graph metadata
        for link in self.user_D.keys():
            d = self.user_D[link]
            if np.sum(d) != 0:
                self.graph.edges[link]['user_D'] = d
                if link in self.sample_path.keys():
                    self.graph.edges[link]['sample_path'] = self.sample_path[link]
            else:
                # Placeholder to avoid key errors
                self.graph.edges[link]['user_D'] = 0
        
        self.num_nodes = self.graph.number_of_nodes()
        self.adjacency_matrix = np.vstack(self.graph.edges())
        # Set node levels
        self.levels = {}
        self.levels['retailer'] = np.array([1])
        self.levels['distributor'] = np.unique(np.hstack(
            [list(self.graph.predecessors(i)) for i in self.levels['retailer']]))
        self.levels['manufacturer'] = np.unique(np.hstack(
            [list(self.graph.predecessors(i)) for i in self.levels['distributor']]))
        self.levels['raw_materials'] = np.unique(np.hstack(
            [list(self.graph.predecessors(i)) for i in self.levels['manufacturer']]))

        self.level_col = {'retailer': 0,
                    'distributor': 1,
                    'manufacturer': 2,
                    'raw_materials': 3}

        self.market = [j for j in self.graph.nodes() if len(list(self.graph.successors(j))) == 0]
        self.distrib = [j for j in self.graph.nodes() if 'C' not in self.graph.nodes[j] and 'I0' in self.graph.nodes[j]]
        self.retail = [j for j in self.graph.nodes() if len(set.intersection(set(self.graph.successors(j)), set(self.market))) > 0]
        self.factory = [j for j in self.graph.nodes() if 'C' in self.graph.nodes[j]]
        self.rawmat = [j for j in self.graph.nodes() if len(list(self.graph.predecessors(j))) == 0]
        self.main_nodes = np.sort(self.distrib + self.factory)
        self.reorder_links = [e for e in self.graph.edges() if 'L' in self.graph.edges[e]] #exclude links to markets (these cannot have lead time 'L')
        self.retail_links = [e for e in self.graph.edges() if 'L' not in self.graph.edges[e]] #links joining retailers to markets
        self.network_links = [e for e in self.graph.edges()] #all links involved in sale in the network

        # check inputs
        assert set(self.graph.nodes()) == set.union(set(self.market),
                                                    set(self.distrib),
                                                    set(self.factory),
                                                    set(self.rawmat)), "The union of market, distribution, factory, and raw material nodes is not equal to the system nodes."
        for j in self.graph.nodes():
            if 'I0' in self.graph.nodes[j]:
                assert self.graph.nodes[j]['I0'] >= 0, "The initial inventory cannot be negative for node {}.".format(j)
            if 'h' in self.graph.nodes[j]:
                assert self.graph.nodes[j]['h'] >= 0, "The inventory holding costs cannot be negative for node {}.".format(j)
            if 'C' in self.graph.nodes[j]:
                assert self.graph.nodes[j]['C'] > 0, "The production capacity must be positive for node {}.".format(j)
            if 'o' in self.graph.nodes[j]:
                assert self.graph.nodes[j]['o'] >= 0, "The operating costs cannot be negative for node {}.".format(j)
            if 'v' in self.graph.nodes[j]:
                assert self.graph.nodes[j]['v'] > 0 and self.graph.nodes[j]['v'] <= 1, "The production yield must be in the range (0, 1] for node {}.".format(j)
        for e in self.graph.edges():
            if 'L' in self.graph.edges[e]:
                assert self.graph.edges[e]['L'] >= 0, "The lead time joining nodes {} cannot be negative.".format(e)
            if 'p' in self.graph.edges[e]:
                assert self.graph.edges[e]['p'] >= 0, "The sales price joining nodes {} cannot be negative.".format(e)
            if 'b' in self.graph.edges[e]:
                assert self.graph.edges[e]['b'] >= 0, "The unfulfilled demand costs joining nodes {} cannot be negative.".format(e)
            if 'g' in self.graph.edges[e]:
                assert self.graph.edges[e]['g'] >= 0, "The pipeline inventory holding costs joining nodes {} cannot be negative.".format(e)
            if 'sample_path' in self.graph.edges[e]:
                assert isinstance(self.graph.edges[e]['sample_path'], bool), "When specifying if a user specified demand joining (retailer, market): {} is sampled from a distribution, sample_path must be a Boolean.".format(e)
            if 'demand_dist' in self.graph.edges[e]:
                dist = self.graph.edges[e]['demand_dist'] #extract distribution
                assert dist.cdf(0,**self.graph.edges[e]['dist_param']), "Wrong parameters passed to the demand distribution joining (retailer, market): {}.".format(e)
        assert self.backlog == False or self.backlog == True, "The backlog parameter must be a boolean."
        assert self.graph.number_of_nodes() >= 2, "The minimum number of nodes is 2. Please try again"
        assert self.alpha>0 and self.alpha<=1, "alpha must be in the range (0, 1]."
        
        # set random generation seed (unless using user demands)
        self.seed(self.seed_int)
        
        # action space (reorder quantities for each node for each supplier; list)
        # An action is defined for every node
        num_reorder_links = len(self.reorder_links) 
        self.lt_max = np.max([self.graph.edges[e]['L'] for e in self.graph.edges() if 'L' in self.graph.edges[e]])
        self.init_inv_max = np.max([self.graph.nodes[j]['I0'] for j in self.graph.nodes() if 'I0' in self.graph.nodes[j]])
        self.capacity_max = np.max([self.graph.nodes[j]['C'] for j in self.graph.nodes() if 'C' in self.graph.nodes[j]])
        self.pipeline_length = sum([self.graph.edges[e]['L']
            for e in self.graph.edges() if 'L' in self.graph.edges[e]])
        self.lead_times = {e: self.graph.edges[e]['L'] 
            for e in self.graph.edges() if 'L' in self.graph.edges[e]}
        self.obs_dim = self.pipeline_length + len(self.main_nodes) + len(self.retail_links)
        # self.pipeline_length = len(self.main_nodes)*(self.lt_max+1)
        self.action_space = gym.spaces.Box(
            low=np.zeros(num_reorder_links),
            high=np.ones(num_reorder_links)*(self.init_inv_max + self.capacity_max*self.num_periods), 
            dtype=np.int32)
        # observation space (total inventory at each node, which is any integer value)
        self.observation_space = gym.spaces.Box(
            low=np.ones(self.obs_dim)*np.iinfo(np.int32).min,
            high=np.ones(self.obs_dim)*np.iinfo(np.int32).max,
            dtype=np.int32)
            # low=-np.ones(self.pipeline_length)*(self.init_inv_max + self.capacity_max*self.num_periods)*10,
            # high=np.ones(self.pipeline_length)*(self.init_inv_max + self.capacity_max*self.num_periods), 
            # dtype=np.int32)

        # intialize
        self.reset()

    def seed(self,seed=None):
        '''
        Set random number generation seed
        '''
        # seed random state
        if seed != None:
            np.random.seed(seed=int(seed))
        
    def _RESET(self):
        '''
        Create and initialize all variables and containers.
        Nomenclature:
            I = On hand inventory at the start of each period at each stage (except last one).
            T = Pipeline inventory at the start of each period at each stage (except last one).
            R = Replenishment order placed at each period at each stage (except last one).
            D = Customer demand at each period (at the retailer)
            S = Sales performed at each period at each stage.
            B = Backlog at each period at each stage.
            LS = Lost sales at each period at each stage.
            P = Total profit at each stage.
        '''
        T = self.num_periods
        J = len(self.main_nodes)
        RM = len(self.retail_links)  # number of retailer-market pairs
        PS = len(self.reorder_links) # number of purchaser-supplier pairs in the network
        SL = len(self.network_links) # number of edges in the network (excluding links form raw material nodes)
        
        # simulation result lists
        self.X=pd.DataFrame(data = np.zeros([T + 1, J]), 
                            columns = self.main_nodes) # inventory at the beginning of each period
        self.Y=pd.DataFrame(data = np.zeros([T + 1, PS]), 
                            columns = pd.MultiIndex.from_tuples(self.reorder_links,
                            names = ['Source','Receiver'])) # pipeline inventory at the beginning of each period
        self.R=pd.DataFrame(data = np.zeros([T, PS]), 
                            columns = pd.MultiIndex.from_tuples(self.reorder_links,
                            names = ['Supplier','Requester'])) # replenishment orders
        self.S=pd.DataFrame(data = np.zeros([T, SL]), 
                            columns = pd.MultiIndex.from_tuples(self.network_links, 
                            names = ['Seller','Purchaser'])) # units sold
        self.D=pd.DataFrame(data = np.zeros([T, RM]), 
                            columns = pd.MultiIndex.from_tuples(self.retail_links, 
                            names = ['Retailer','Market'])) # demand at retailers
        self.U=pd.DataFrame(data = np.zeros([T, RM]), 
                            columns = pd.MultiIndex.from_tuples(self.retail_links, 
                            names = ['Retailer','Market'])) # unfulfilled demand for each market - retailer pair
        self.P=pd.DataFrame(data = np.zeros([T, J]), 
                            columns = self.main_nodes) # profit at each node
        
        # initializetion
        self.period = 0 # initialize time
        for j in self.main_nodes:
            self.X.loc[0,j]=self.graph.nodes[j]['I0'] # initial inventory
        self.Y.loc[0,:]=np.zeros(PS) # initial pipeline inventory
        self.action_log = np.zeros([T, PS])

        # set state
        self._update_state()
        
        return self.state

    def _update_state(self):
        # State is a concatenation of demand, inventory, and pipeline at each time step
        demand = np.hstack([self.D[d].iloc[self.period] for d in self.retail_links])
        inventory = np.hstack([self.X[n].iloc[self.period] for n in self.main_nodes])

        # Pipeline values won't be of proper dimension if current
        # current period < lead time. We need to add 0's as padding.
        if self.period == 0:
            _pipeline = [[self.Y[k].iloc[0]]
                for k, v in self.lead_times.items()]
        else:
            _pipeline = [self.Y[k].iloc[max(self.period-v,0):self.period].values
                for k, v in self.lead_times.items()]
        pipeline = []
        for p, v in zip(_pipeline, self.lead_times.values()):
            if v == 0:
                continue
            if len(p) <= v:
                pipe = np.zeros(v)
                pipe[-len(p):] += p
            pipeline.append(pipe)
        pipeline = np.hstack(pipeline)
        self.state = np.hstack([demand, inventory, pipeline])
    
    def _STEP(self, action):
        '''
        Take a step in time in the multiperiod inventory management problem.
        action = number of units to request from each supplier.
            dictionary: keys are (supplier, purchaser) tuples
                        values are number of units requested from supplier
                        dimension = len(reorder_links) (number of edges joining all nodes, 
                                                        except market nodes)
        '''
        t = self.period
        if type(action) != dict: # convert to dict if a list was given
            action = {key: action[i] for i, key in enumerate(self.reorder_links)}
        
        # Place Orders
        for key in action.keys():
            request = round(max(action[key],0)) # force to integer value
            supplier = key[0]
            purchaser = key[1]
            if supplier in self.rawmat:
                self.R.loc[t,(supplier, purchaser)] = request # accept request since supply is unlimited
                self.S.loc[t,(supplier, purchaser)] = request
            elif supplier in self.distrib:
                X_supplier = self.X.loc[t,supplier] # request limited by available inventory at beginning of period
                self.R.loc[t,(supplier, purchaser)] = min(request, X_supplier)
                self.S.loc[t,(supplier, purchaser)] = min(request, X_supplier)
            elif supplier in self.factory:
                C = self.graph.nodes[supplier]['C'] # supplier capacity
                v = self.graph.nodes[supplier]['v'] # supplier yield
                X_supplier = self.X.loc[t,supplier] # on-hand inventory at beginning of period
                self.R.loc[t,(supplier, purchaser)] = min(request, C, v*X_supplier)
                self.S.loc[t,(supplier, purchaser)] = min(request, C, v*X_supplier)
            
        #Receive deliveries and update inventories
        for j in self.main_nodes:
            #update pipeline inventories
            incoming = []
            for k in self.graph.predecessors(j):
                L = self.graph.edges[(k,j)]['L'] #extract lead time
                if t - L >= 0: #check if delivery has arrived
                    delivery = self.R.loc[t-L,(k,j)]
                else:
                    delivery = 0
                incoming += [delivery] #update incoming material
                self.Y.loc[t+1,(k,j)] = self.Y.loc[t,(k,j)] - delivery + self.R.loc[t,(k,j)]

            #update on-hand inventory
            if 'v' in self.graph.nodes[j]: #extract production yield
                v = self.graph.nodes[j]['v']
            else:
                v = 1
            outgoing = 1/v * np.sum([self.S.loc[t,(j,k)] for k in self.graph.successors(j)]) #consumed inventory (for requests placed)
            self.X.loc[t+1,j] = self.X.loc[t,j] + np.sum(incoming) - outgoing
            
        # demand is realized
        for j in self.retail:
            for k in self.market:
                #read user specified demand. if all zeros, use demand_dist instead.
                Demand = self.graph.edges[(j,k)]['user_D']
                if np.sum(Demand) > 0:
                    self.D.loc[t,(j,k)] = Demand[t]
                else:
                    Demand = self.graph.edges[(j,k)]['demand_dist']
                    self.D.loc[t,(j,k)] = Demand.rvs(
                        **self.graph.edges[(j,k)]['dist_param'])
                if self.backlog and t >= 1:
                    D = self.D.loc[t,(j,k)] + self.U.loc[t-1,(j,k)]
                else:
                    D = self.D.loc[t,(j,k)]
                #satisfy demand up to available level
                X_retail = self.X.loc[t+1,j] #get inventory at retail before demand was realized
                self.S.loc[t,(j,k)] = min(D, X_retail) #perform sale
                self.X.loc[t+1,j] -= self.S.loc[t,(j,k)] #update inventory
                self.U.loc[t,(j,k)] = D - self.S.loc[t,(j,k)] #update unfulfilled orders

        # calculate profit
        for j in self.main_nodes:
            a = self.alpha
            SR = np.sum([self.graph.edges[(j,k)]['p'] * self.S.loc[t,(j,k)] for k in self.graph.successors(j)]) #sales revenue
            PC = np.sum([self.graph.edges[(k,j)]['p'] * self.R.loc[t,(k,j)] for k in self.graph.predecessors(j)]) #purchasing costs
            if j not in self.rawmat:
                HC = self.graph.nodes[j]['h'] * self.X.loc[t+1,j] + np.sum([self.graph.edges[(k,j)]['g'] * self.Y.loc[t+1,(k,j)] for k in self.graph.predecessors(j)]) #holding costs
            else:
                HC = 0
            if j in self.factory:
                OC = self.graph.nodes[j]['o'] / self.graph.nodes[j]['v'] * np.sum([self.S.loc[t,(j,k)] for k in self.graph.successors(j)]) #operating costs
            else:
                OC = 0
            if j in self.retail:
                UP = np.sum([self.graph.edges[(j,k)]['b'] * self.U.loc[t,(j,k)] for k in self.graph.successors(j)]) #unfulfilled penalty
            else:
                UP = 0
            self.P.loc[t,j] = a**t * (SR - PC - OC - HC - UP)
        
        # update period
        self.period += 1

        # set reward (profit from current timestep)
        reward = self.P.loc[t,:].sum()
        
        # determine if simulation should terminate
        if self.period >= self.num_periods:
            done = True
        else:
            done = False
            # update stae
            self._update_state()

        return self.state, reward, done, {}
    
    def sample_action(self):
        '''
        Generate an action by sampling from the action_space
        '''
        return self.action_space.sample()

    def step(self, action):
        return self._STEP(action)

    def reset(self):
        return self._RESET()

    def plot_network(self):
        colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
        adjacency_matrix = np.vstack(self.graph.edges())
        # Set level colors
        level_col = {'retailer': 0,
                    'distributor': 1,
                    'manufacturer': 2,
                    'raw_materials': 3}

        max_density = np.max([len(v) for v in self.levels.values()])
        node_coords = {}
        node_num = 1
        plt.figure(figsize=(12,8))
        for i, (level, nodes) in enumerate(self.levels.items()):
            n = len(nodes)
            node_y = max_density / 2 if n == 1 else np.linspace(0, max_density, n)
            node_y = np.atleast_1d(node_y)
            plt.scatter(np.repeat(i, n), node_y, label=level, s=50)
            for y in node_y:
                plt.annotate(r'$N_{}$'.format(node_num), xy=(i, y+0.05))
                node_coords[node_num] = (i, y)
                node_num += 1

        # Draw edges
        for node_num, v in node_coords.items():
            x, y = v
            sinks = adjacency_matrix[np.where(adjacency_matrix[:, 0]==node_num)][:, 1]
            for s in sinks:
                try:
                    sink_coord = node_coords[s]
                except KeyError:
                    continue
                for k, n in self.levels.items():
                    if node_num in n:
                        color = colors[level_col[k]]
                x_ = np.hstack([x, sink_coord[0]])
                y_ = np.hstack([y, sink_coord[1]])
                plt.plot(x_, y_, color=color)

        plt.ylabel('Node')
        plt.yticks([0], [''])
        plt.xlabel('Level')
        plt.xticks(np.arange(len(self.levels)), [k for k in self.levels.keys()])
        plt.show()
        
class NetInvMgmtBacklogEnv(NetInvMgmtMasterEnv):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
class NetInvMgmtLostSalesEnv(NetInvMgmtMasterEnv):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.backlog = False


from or_gym.envs.classic_or.knapsack import *
from or_gym.envs.classic_or.binpacking import *
from or_gym.envs.classic_or.vmpacking import *
from or_gym.envs.classic_or.tsp import *
from or_gym.envs.classic_or.vehicle_routing import VehicleRoutingEnv
from or_gym.envs.classic_or.newsvendor import NewsvendorEnv


'''
Example taken from Balaji et al.
Paper: https://arxiv.org/abs/1911.10641
GitHub: https://github.com/awslabs/or-rl-benchmarks
'''
import gym
from gym import spaces
import or_gym
from or_gym.utils import assign_env_config
import random
import numpy as np
from scipy.stats import truncnorm

class VehicleRoutingEnv(gym.Env):
    '''
    Dynamic Vehicle Routing Problem

    This environment simulates a driver working with a food delivery app
    to move through a city, accept orders, pick them up from restaurants,
    and deliver them to waiting customers. Each order has a specific
    delivery value, restaurant, and delivery location, all of which are 
    known by the driver before he accepts the order. After accepting, the
    driver must navigate to the restaurant to collect the order and then
    deliver it. If an order isn't accepted, it may be taken by another
    driver. Additionally, the driver has 60 minutes to make a delivery
    from the time an order is created. 
    The city is represented as a grid with different zones that have
    different statistics for order creation and value. At each time step,
    new orders are created with a fixed probability unique to each zone.
    The driver's vehicle also has a finite capacity limiting the number of
    orders he can carry at a given time, although there is no limit on the
    number of accepted orders.
    The driver receives a penalty for time and distance spent during travel,
    but receives rewards for accepting and delivering orders.

    Observation:
        Type: Box
        State Vector: S = (p, h, c, l, w, e, v)
        p = pickup location
        h = driver's current position
        c = remaining vehicle capacity
        l = order location
        w = order status (open, accepted, picked up, delivered/inactive)
        e = time elapsed since order generation
        v = order value

    Action: 
        Type: Discrete
        0 = wait
        1:max_orders = accept order
        max_orders:2*max_orders = pickup order
        2*max_orders:3*max_orders = deliver order
        3*max_orders:3*max_orders + n_restaurants = go to restaurant

        Action masking is available for this environment. Set mask=True
        in the env_config dictionary.

    Reward:
        The agent recieves 1/3 of the order value for accepting an order,
        picking it up, and delivering the order. The cost is comprised of
        three elements: delivery time, delivery distance, and cost of failure
        (if the driver does not deliver the item). 
    
    Starting State:
        Restaurant and driver locations are randomized at the start of each
        episode. New orders are generated according to the order probability.

    Episode Terimantion:
        Episode termination occurs when the total time has elapsed.
    '''
    def __init__(self, *args, **kwargs):
        self.n_restaurants = 2
        self.max_orders = 10
        self.order_prob = 0.5
        self.vehicle_capacity = 4
        self.grid = (5, 5)
        self.order_promise = 60
        self.order_timeout_prob = 0.15
        self.num_zones = 4
        self.order_probs_per_zone = [0.1, 0.5, 0.3, 0.1]
        self.order_reward_min = [8, 5, 2, 1]
        self.order_reward_max = [12, 8, 5, 3]
        self.half_norm_scale_reward_per_zone = [0.5, 0.5, 0.5, 0.5]
        self.penalty_per_timestep = 0.1
        self.penalty_per_move = 0.1
        self.order_miss_penalty = 50
        self.step_limit = 1000
        self.mask = False
        self.info = {}
        
        assign_env_config(self, kwargs)
        self._order_nums = np.arange(self.max_orders)
        self.loc_permutations = [(x, y) for x in range(self.grid[0])
            for y in range(self.grid[1])]
        self.action_dim = 1 + 3 * self.max_orders + self.n_restaurants
        self.obs_dim = 2 * self.n_restaurants + 4 + 6 * self.max_orders
        box_low = np.zeros(self.obs_dim)
        box_high = np.hstack([
            np.repeat(
                max(self.grid), 2 * self.n_restaurants + 2), # Locations 0-5
            np.repeat(self.vehicle_capacity, 2), # Vehicle capacities 6-7
            np.tile(np.hstack([4, self.n_restaurants, self.grid, 
            self.order_promise, max(self.order_reward_max)]), self.max_orders)
        ])

        if self.mask:
            self.observation_space = spaces.Dict({
                'action_mask': spaces.Box(
                    low=np.zeros(self.action_dim),
                    high=np.ones(self.action_dim),
                    dtype=np.uint8),
                'avail_actions': spaces.Box(
                    low=np.zeros(self.action_dim),
                    high=np.ones(self.action_dim),
                    dtype=np.uint8),
                'state': spaces.Box(
                    low=box_low,
                    high=box_high,
                    dtype=np.float16)
            })
        else:
            self.observation_space = spaces.Box(
                low=box_low,
                high=box_high,
                dtype=np.float16)
                    
        self.action_space = spaces.Discrete(self.action_dim)
        
        self.reset()
        
    def _STEP(self, action):
        done = False
        self.reward = 0
        self.late_penalty = 0
        
        if action == 0:
            self.wait(action)
        elif action <= self.max_orders:
            self.accept_order(action)
        elif action <= 2 * self.max_orders:
            self.pickup_order(action)
        elif action <= 3 * self.max_orders:
            self.deliver_order(action)
        elif action <= 3 * self.max_orders + self.n_restaurants:
            self.return_to_restaurant(action)
        else:
            raise Exception(f"Selected action ({action}) outside of action space.")
            
        self.state = self._update_state()
        
        self.step_count += 1
        if self.step_count >= self.step_limit:
            done = True
            
        return self.state, self.reward, done, self.info
        
    def wait(self, action):
        # Do nothing
        pass
        
    def accept_order(self, action):
        # Accept order denoted by action
        order_idx = action - 1
        if order_idx not in self.order_dict.keys():
            # Invalid action, do nothing
            pass
        elif self.order_dict[order_idx] == 1:
            self.order_dict[order_idx] = 2
            self.reward += self.order_dict[order_idx]['Value'] / 3
        
    def pickup_order(self, action):
        order_idx = action - self.max_orders - 1
        if order_idx not in self.order_dict.keys():
            # Invalid action, do nothing
            pass
        else:
            restaurant = self.order_dict[order_idx]['RestaurantID']
            restaurant_loc = self.restaurant_loc[restaurant]
            self._go_to_destination(restaurant_loc)
            self.reward -= self.penalty_per_move
            # Movement and pickup can occur during same time step
            if self.order_dict[order_idx]['Status'] == 2 and self.driver_loc == restaurant_loc:
                if self.vehicle_load < self.vehicle_capacity:
                    self.order_dict[order_idx]['Status'] = 3
                    self.vehicle_capacity += 1
                    self.reward += self.order_dict[order_idx]['Value'] / 3
        
    def deliver_order(self, action):
        order_idx = action - 2 * self.max_orders - 1
        if order_idx not in self.order_dict.keys():
            # Invalid action, do nothing
            pass
        else:
            order_loc = self.order_dict[order_idx]['DeliveryLoc']
            self._go_to_destination(order_loc)
            self.reward -= self.penalty_per_move
            # Can deliver multiple orders simultaneously
            for k, v in self.order_dict.items():
                if v['Status'] == 3 and v['DeliveryLoc'] == self.driver_loc:
                    if v['Time'] <= self.order_promise:
                        self.reward = v['Value'] / 3
                    self.vehicle_load -= 1
                    v['Status'] = 4 # Delivered

    def return_to_restaurant(self, action):
        restaurant = action - 3 * self.max_orders - 1
        restaurant_loc = self.restaurant_loc[restaurant]
        self._go_to_destination(restaurant_loc)
        self.reward -= self.penalty_per_move
    
    def _update_orders(self):
        self._update_order_times()
        self._remove_orders()
        self._generate_orders()
        
    def _remove_orders(self):
        # Remove orders if they're over due
        orders_to_delete = []
        for k, v in self.order_dict.items():
            if v['Time'] >= self.order_promise:
                if v['Status'] >= 2:
                    # Apply penalty and remove associated rewards
                    self.reward -= (self.order_miss_penalty +
                        v['Value'] * (v['Status']==2)/3 +
                        v['Value'] * (v['Status']==3) * 2/3)
                    self.late_penalty += self.order_miss_penalty
                if v['Status'] == 3:
                    self.vehicle_capacity -= 1
                orders_to_delete.append(k)
                
            elif v['Status'] == 4:
                orders_to_delete.append(k)
                
            # Probabalistically remove open orders
            elif v['Status'] == 1 and np.random.random() < self.order_timeout_prob:
                orders_to_delete.append(k)
                
        for k in orders_to_delete:
            del self.order_dict[k]
    
    def _update_state(self):
        self._update_orders()
        order_array = np.zeros((self.max_orders, 6)) # Placeholder for order data
        try:
            order_data = np.hstack([v1 for v in self.order_dict.values() 
                for v1 in v.values()]).reshape(-1, 7)
            order_array[order_data[:, 0].astype(int)] += order_data[:, 1:]
        except ValueError:
            # Occurs when order_data is empty
            pass
        state = np.hstack([
            np.hstack(self.restaurant_loc),
            np.hstack(self.driver_loc),
            np.hstack([self.vehicle_load, self.vehicle_capacity]),
            order_array.flatten()
        ])
        if self.mask:
            action_mask = self._update_mask(state)
            state = {
                'state': state,
                'action_mask': action_mask,
                'avail_actions': np.ones(self.action_dim)
            }
        return state

    def _update_mask(self, state):
        action_mask = np.zeros(self.action_dim)
        # Wait and return to restaurant are always allowed
        action_mask[0] = 1
        action_mask[(3 * self.max_orders + 1):(3 * self.max_orders + self.n_restaurants + 1)] = 1

        for k, v in self.order_dict.items():
            status = v['Status']
            # Allow accepting an open order
            if status == 1:
                action_mask[k + 1] = 1
            # Allow navigating to accepted order for pickup
            elif status == 2 and self.vehicle_load < self.vehicle_capacity:
                action_mask[k + self.max_orders + 1] = 1
            # Allow delivery of picked up order
            elif status == 3:
                action_mask[k + 2 * self.max_orders + 1] = 1

        return action_mask

    def _RESET(self):
        self.step_count = 0
        self.vehicle_load = 0
        self.randomize_locations()
        self.zone_loc = self._get_zones()
        self.order_dict = {}
        self.state = self._update_state()
        return self.state
    
    def _update_order_times(self):
        for k, v in self.order_dict.items():
            if v['Status'] >= 1:
                v['Time'] += 1
    
    def _generate_orders(self):
        open_slots = self._order_nums[~np.isin(self._order_nums,
            np.array([k for k in self.order_dict.keys()]))]
        try:
            order_num = open_slots.min()
        except ValueError:
            pass
        for n in open_slots:
            # Probabalistically create a new order
            if np.random.random() < self.order_prob:
                zone = np.random.choice(self.num_zones, p=self.order_probs_per_zone)
                order = self._get_order_from_zone(zone, order_num)
                self.order_dict[order_num] = order
                order_num += 1
            
    def _get_order_from_zone(self, zone, n):
        delivery_loc = random.choice(self.zone_loc[zone])
        restaurant_idx = np.random.choice(self.n_restaurants)
        value = truncnorm.rvs(0, 
            (self.order_reward_max[zone] - self.order_reward_min[zone])
                              /self.half_norm_scale_reward_per_zone[zone],
            self.order_reward_min[zone],
            self.half_norm_scale_reward_per_zone[zone])
        return {'Number': n,
         'Status': 1,
         'RestaurantID': restaurant_idx,
         'DeliveryLoc': delivery_loc,
         'Time': 0,
         'Value': value}
        
    def randomize_locations(self):
        self._place_restaurants()
        self._place_driver()
        
    def _place_restaurants(self):
        self.restaurant_loc = random.sample(self.loc_permutations,
            self.n_restaurants)
    
    def _place_driver(self):
        self.driver_loc = list(random.sample(self.loc_permutations, 1)[0])
        
    def _move_driver(self, direction):
        if direction is None:
            return None
        # Receives direction from routing function
        if direction == 0: # Up
            self.driver_loc[1] += 1
        elif direction == 1: # Down
            self.driver_loc[1] -= 1
        elif direction == 2: # Right
            self.driver_loc[0] += 1
        elif direction == 3: # Left
            self.driver_loc[0] -= 1
        # Check boundaries
        if self.driver_loc[0] > self.grid[0]:
            self.driver_loc[0] = self.grid[0]
        if self.driver_loc[0] < 0:
            self.driver_loc[0] = 0
        if self.driver_loc[1] > self.grid[1]:
            self.driver_loc[1] = self.grid[1]
        if self.driver_loc[1] < 0:
            self.driver_loc[1] = 0
            
    def _go_to_destination(self, destination):
        # Automatically selects direction based on starting location and
        # destination. 
        # 0 -> Up; 1 -> Down; 2 -> Right; 3 -> Left
        x_diff = self.driver_loc[0] - destination[0]
        y_diff = self.driver_loc[1] - destination[1]
        if abs(x_diff) >= abs(y_diff):
            if x_diff > 0:
                direction = 2
            elif x_diff < 0:
                direction = 3
            elif abs(x_diff) == abs(y_diff): # 0 == 0
                # Do nothing
                direction = None
        else:
            if y_diff > 0:
                direction = 0
            elif y_diff < 0:
                direction = 1
        
        self._move_driver(direction)
            
    def _get_num_spaces_per_zone(self):
        total_spaces = self.grid[0] * self.grid[1]
        spaces_per_zone = np.array([np.floor(total_spaces / self.num_zones) 
            for i in range(self.num_zones)])
        for i in range(total_spaces % self.num_zones):
            spaces_per_zone[i] += 1
        return spaces_per_zone.astype(int)
    
    def _get_zones(self):
        # Slices the grid into zones by row
        spaces_per_zone = self._get_num_spaces_per_zone()
        zones = {}
        for i, n in enumerate(spaces_per_zone):
            x = sum(spaces_per_zone[:i])
            zones[i] = self.loc_permutations[x:x+n]
            
        zones = self._remove_restaurants_from_zone_locs(zones)
        return zones
    
    def _remove_restaurants_from_zone_locs(self, zones):
        for k, v in zones.items():
            for r in self.restaurant_loc:
                try:
                    loc_to_remove = v.index(r)
                    del zones[k][loc_to_remove]
                except ValueError:
                    pass
        return zones

    def step(self, action):
        return self._STEP(action)

    def reset(self):
        return self._RESET()


'''
Example taken from Balaji et al.
Paper: https://arxiv.org/abs/1911.10641
GitHub: https://github.com/awslabs/or-rl-benchmarks
'''
import numpy as np
import gym
from gym import spaces, logger
from gym.utils import seeding
from or_gym.utils import assign_env_config
import copy

BIG_NEG_REWARD = -100
BIG_POS_REWARD = 10

class BinPackingEnv(gym.Env):
    '''
    Small Bin Packing with Bounded Waste
    Env Registration: BinPacking-v0

    The Bin Packing Problem (BPP) is a combinatorial optimization problem which
    requires the user to select from a range of goods of different values and
    weights in order to maximize the value of the selected items within a 
    given weight limit. This version is online meaning each item is randonly
    presented to the algorithm one at a time, at which point the algorithm 
    can either accept or reject the item. After seeing a fixed number of 
    items are shown, the episode terminates. If the weight limit is reached
    before the episode ends, then it terminates early.

    Observation:
        If mask == False:
            Type: Discrete
            0 - bin_capacity: Count of bins at a given level h
            -1: Current item size
        if mask == True:
            Type: Dict
            'state': vector of bins where 0 to bin capacity is the count of
                bins at that load level h and the last entry is the current
                item size.
            'action_mask': binary vector where 0 indicates infeasible
                actions and 1 feasible actions.
            'avail_actions': vector of values to be combined with mask.

    Actions:
        Type: Discrete
        0: Open a new bin and place item into bin
        1+: Attempt to place item into bin at the corresponding level

    Reward:
        Negative of the waste, which is the difference between the current
        size and excess space of the bin.

    Starting State:
        No available bins and random starting item
        
    Episode Termination:
        When invalid action is selected (e.g. attempt to place item in non-existent
        bin), bin limits are exceeded, or step limit is reached.
    '''
    def __init__(self, *args, **kwargs):
        self.bin_capacity = 9
        self.item_sizes = [2, 3]
        self.item_probs = [0.8, 0.2]
        self.step_count = 0
        self.step_limit = 100
        self.mask = False
        assign_env_config(self, kwargs)
        self._build_obs_space()
        self._check_settings()
        self.seed()
        self.state = self.reset()
        
    def _STEP(self, action):
        done = False
        if action >= self.bin_capacity:
            raise ValueError('{} is an invalid action. Must be between {} and {}'.format(
                action, 0, self.bin_capacity))
        elif action > (self.bin_capacity - self.item_size):
            # Bin overflows
            reward = BIG_NEG_REWARD - self.waste
            done = True
        elif action == 0:
            # Create new bin
            self.bin_levels[self.item_size] += 1
            self.waste = self.bin_capacity - self.item_size
            reward = -1 * self.waste
        elif self.bin_levels[action] == 0:
            # Can't insert item into non-existent bin
            reward = BIG_NEG_REWARD - self.waste
            done = True
        else:
            if action + self.item_size == self.bin_capacity:
                self.num_full_bins += 1
            else:
                self.bin_levels[action + self.item_size] += 1
            self.waste = -self.item_size
            reward = -1 * self.waste
            
            self.bin_levels[action] -= 1
        
        self.total_reward += reward
        
        if self.step_count >= self.step_limit:
            done = True
        
        self.state = self._update_state()
        
        return self.state, reward, done, {}

    def _update_state(self):
        self.item_size = self.get_item()
        state = np.array(self.bin_levels + [self.item_size])
        if self.mask:
            state_dict = {
                'state': state,
                'avail_actions': np.ones(self.bin_capacity)}
            # Mask actions for closed bins
            mask = np.ones(self.bin_capacity) * np.array(state[:-1])
            # Mask actions where packing would exceed capacity
            overflow = self.bin_capacity - self.item_size
            mask[overflow+1:] = 0
            # Ensure open new bin is available
            mask[0] = 1
            state_dict['action_mask'] = mask
            return state_dict
        else:
            return state
    
    def get_item(self):
        return np.random.choice(self.item_sizes, p=self.item_probs)
        
    def sample_action(self):
        return self.action_space.sample()
    
    def _RESET(self):
        self.current_weight = 0
        self.step_count = 0        
        self.num_full_bins = 0
        self.total_reward = 0
        self.waste = 0
        self.bin_levels = [0] * self.bin_capacity
        self.item_size = self.get_item()
        self.state = self._update_state()
        return self.state

    def _build_obs_space(self):
        if self.mask:
            self.observation_space = spaces.Dict({
                'action_mask': spaces.Box(0, 1, 
                    shape=(self.bin_capacity,),
                    dtype=np.uint8),
                'avail_actions': spaces.Box(0, 1, 
                    shape=(self.bin_capacity,),
                    dtype=np.uint8),
                'state': spaces.Box(
                    low=np.array([0] * (1 + self.bin_capacity)),
                    high=np.array([self.step_limit] * self.bin_capacity +
                        [max(self.item_sizes)]),
                    dtype=np.uint32)
            })
        else:
            self.observation_space = spaces.Box(
                low=np.array([0] * (1 + self.bin_capacity)),
                high=np.array([self.step_limit] * self.bin_capacity + 
                    [max(self.item_sizes)]),
                dtype=np.uint32)
        
        self.action_space = spaces.Discrete(self.bin_capacity)

    def _check_settings(self):
        # Ensure setting sizes and probs are correct at initialization
        assert sum(self.item_probs) == 1, 'Item probabilities do not sum to 1.'
        assert len(self.item_probs) == len(self.item_sizes), \
            'Dimension mismatch between item probabilities' + \
                ' ({}) and sizes ({})'.format(
                len(self.item_probs), len(self.item_sizes))

    def reset(self):
        return self._RESET()

    def step(self, action):
        return self._STEP(action)

class BinPackingLW1(BinPackingEnv):
    '''
    Large Bin Packing Probem with Bounded Waste
    Env Registration: BinPacking-v1
    '''
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.bin_capacity = 100
        self.item_probs = [0.14, 0.1, 0.06, 0.13, 0.11, 0.13, 0.03, 0.11, 0.19]
        self.item_sizes = np.arange(1, 10)
        self.step_limit = 1000
        assign_env_config(self, kwargs)
        self._build_obs_space()
        self._check_settings()
        self.seed()
        self.state = self.reset()

class BinPackingPP0(BinPackingEnv):
    '''
    Small Perfectly Packable Bin Packing with Linear Waste
    Env Registration: BinPacking-v2
    '''
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.item_probs = [0.75, 0.25]
        assign_env_config(self, kwargs)
        self._build_obs_space()
        self._check_settings()
        self.seed()
        self.state = self.reset()

class BinPackingPP1(BinPackingPP0):
    '''
    Large Bin Packing Probem with Bounded Waste
    Env Registration: BinPacking-v3
    '''
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.bin_capacity = 100
        self.item_probs = [0.06, 0.11, 0.11, 0.22, 0, 0.11, 0.06, 0, 0.33]
        self.item_sizes = np.arange(1, 10)
        self.step_limit = 1000
        assign_env_config(self, kwargs)
        self._build_obs_space()
        self._check_settings()
        self.seed()
        self.state = self.reset()

class BinPackingBW0(BinPackingEnv):
    '''
    Small Perfectly Packable Bin Packing Problem with Bounded Waste
    Env Registration: BinPacking-v4
    '''
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.item_probs = [0.5, 0.5]
        assign_env_config(self, kwargs)
        self._build_obs_space()
        self._check_settings()
        self.seed()
        self.state = self.reset()

class BinPackingBW1(BinPackingBW0):
    '''
    Large Perfectly Packable Bin Packing Problem with Bounded Waste
    Env Registration: BinPacking-v5
    '''
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.bin_capacity = 100
        self.item_probs = [0, 0, 0, 1/3, 0, 0, 0, 0, 2/3]
        self.item_sizes = np.arange(1, 10)
        self.step_limit = 1000
        assign_env_config(self, kwargs)
        self._build_obs_space()
        self._check_settings()
        self.seed()
        self.state = self.reset()


from or_gym.envs.supply_chain.inventory_management import InvManagementBacklogEnv
from or_gym.envs.supply_chain.inventory_management import InvManagementLostSalesEnv
from or_gym.envs.supply_chain.network_management import NetInvMgmtBacklogEnv
from or_gym.envs.supply_chain.network_management import NetInvMgmtLostSalesEnv


import numpy as np
import gym
from gym import spaces, logger
from gym.utils import seeding
from or_gym.utils import assign_env_config
import copy

class VMPackingEnv(gym.Env):
    '''
    Online VM Packing Problem

    The VM Packing Problem (VMPP) is a combinatorial optimization problem which
    requires the user to select from a series of physical machines (PM's) to
    send a virtual machine process to. Each VM process is characterized by
    two values, the memory and compute of the process. These are normalized
    by the PM capacities to range between 0-1. 

    Observation:
        Type: Tuple, Discrete
        [0][:, 0]: Binary indicator for open PM's
        [0][:, 1]: CPU load of PM's
        [0][:, 2]: Memory load of PM's
        [1][0]: Current CPU demand
        [1][1]: Current memory demand

    Actions:
        Type: Discrete
        Integer of PM number to send VM to that PM

    Reward:
        Negative of the waste, which is the difference between the current
        size and excess space on the PM.

    Starting State:
        No open PM's and random starting item
        
    Episode Termination:
        When invalid action is selected, attempt to overload VM, or step
        limit is reached.
    '''
    def __init__(self, *args, **kwargs):
        self.cpu_capacity = 1
        self.mem_capacity = 1
        self.t_interval = 20
        self.tol = 1e-5
        self.step_limit = int(60 * 24 / self.t_interval)
        self.n_pms = 50
        self.load_idx = np.array([1, 2])
        self.seed = 0
        self.mask = True
        assign_env_config(self, kwargs)
        self.action_space = spaces.Discrete(self.n_pms)

        if self.mask:
            self.observation_space = spaces.Dict({
                "action_mask": spaces.Box(0, 1, shape=(self.n_pms,)),
                "avail_actions": spaces.Box(0, 1, shape=(self.n_pms,)),
                "state": spaces.Box(0, 1, shape=(self.n_pms+1, 3))
            })
        else:
            self.observation_space = spaces.Box(0, 1, shape=(self.n_pms+1, 3))
        self.reset()
        
    def _RESET(self):
        self.demand = self.generate_demand()
        self.current_step = 0
        self.state = {
            "action_mask": np.ones(self.n_pms),
            "avail_actions": np.ones(self.n_pms),
            "state": np.vstack([
                np.zeros((self.n_pms, 3)),
                self.demand[self.current_step]])
        }
        self.assignment = {}
        return self.state
    
    def _STEP(self, action):
        done = False
        pm_state = self.state["state"][:-1]
        demand = self.state["state"][-1, 1:]
        
        if action < 0 or action >= self.n_pms:
            raise ValueError("Invalid action: {}".format(action))
            
        elif any(pm_state[action, 1:] + demand > 1 + self.tol):
            # Demand doesn't fit into PM
            reward = -1000
            done = True
        else:
            if pm_state[action, 0] == 0:
                # Open PM if closed
                pm_state[action, 0] = 1
            pm_state[action, self.load_idx] += demand
            reward = np.sum(pm_state[:, 0] * (pm_state[:,1:].sum(axis=1) - 2))
            self.assignment[self.current_step] = action
            
        self.current_step += 1
        if self.current_step >= self.step_limit:
            done = True
        self.update_state(pm_state)
        return self.state, reward, done, {}
    
    def update_state(self, pm_state):
        # Make action selection impossible if the PM would exceed capacity
        step = self.current_step if self.current_step < self.step_limit else self.step_limit-1
        data_center = np.vstack([pm_state, self.demand[step]])
        data_center = np.where(data_center>1,1,data_center) # Fix rounding errors
        self.state["state"] = data_center
        self.state["action_mask"] = np.ones(self.n_pms)
        self.state["avail_actions"] = np.ones(self.n_pms)
        if self.mask:
            action_mask = (pm_state[:, 1:] + self.demand[step, 1:]) <= 1
            self.state["action_mask"] = (action_mask.sum(axis=1)==2).astype(int)

    def sample_action(self):
        return self.action_space.sample()

    def generate_demand(self):
        n = self.step_limit
        # From Azure data
        mem_probs = np.array([0.12 , 0.165, 0.328, 0.287, 0.064, 0.036])
        mem_bins = np.array([0.02857143, 0.05714286, 0.11428571, 0.45714286, 0.91428571,
           1.]) # Normalized bin sizes
        mu_cpu = 16.08
        sigma_cpu = 1.26
        cpu_demand = np.random.normal(loc=mu_cpu, scale=sigma_cpu, size=n)
        cpu_demand = np.where(cpu_demand<=0, mu_cpu, cpu_demand) # Ensure demand isn't negative
        mem_demand = np.random.choice(mem_bins, p=mem_probs, size=n)
        return np.vstack([np.arange(n)/n, cpu_demand/100, mem_demand]).T

    def step(self, action):
        return self._STEP(action)

    def reset(self):
        return self._RESET()

class TempVMPackingEnv(VMPackingEnv):
    '''
    Online Temporary VM Packing Problem

    The VM Packing Problem (VMPP) is a combinatorial optimization problem which
    requires the user to select from a series of physical machines (PM's) to
    send a virtual machine process to. Each VM process is characterized by
    two values, the memory and compute of the process. These are normalized
    by the PM capacities to range between 0-1. 

    Observation:
        Type: Tuple, Discrete
        [0][:, 0]: Binary indicator for open PM's
        [0][:, 1]: CPU load of PM's
        [0][:, 2]: Memory load of PM's
        [1][0]: Current CPU demand
        [1][1]: Current memory demand

    Actions:
        Type: Discrete
        Integer of PM number to send VM to that PM

    Reward:
        Negative of the waste, which is the difference between the current
        size and excess space on the PM.

    Starting State:
        No open PM's and random starting item
        
    Episode Termination:
        When invalid action is selected, attempt to overload VM, or step
        limit is reached.
    '''
    def __init__(self, *args, **kwargs):
        super().__init__()       
        self.state = self.reset()

    def step(self, action):
        done = False
        pm_state = self.state["state"][:-1]
        demand = self.state["state"][-1, 1:]
        
        if action < 0 or action >= self.n_pms:
            raise ValueError("Invalid action: {}".format(action))
            
        elif any(pm_state[action, 1:] + demand > 1 + self.tol):
            # Demand doesn't fit into PM
            reward = -1000
            done = True
        else:
            if pm_state[action, 0] == 0:
                # Open PM if closed
                pm_state[action, 0] = 1
            pm_state[action, self.load_idx] += demand
            reward = np.sum(pm_state[:, 0] * (pm_state[:,1:].sum(axis=1) - 2))
            self.assignment[self.current_step] = action

        # Remove processes
        if self.current_step in self.durations.values():
            for process in self.durations.keys():
                # Remove process from PM
                if self.durations[process] == self.current_step:
                    pm = self.assignment[process] # Find PM where process was assigned
                    pm_state[pm, self.load_idx] -= self.demand[process]
                    # Shut down PM's if state is 0
                    if pm_state[pm, self.load_idx].sum() == 0:
                        pm_state[pm, 0] = 0
            
        self.current_step += 1
        if self.current_step >= self.step_limit:
            done = True
        self.update_state(pm_state)
        return self.state, reward, done, {}
    
    def update_state(self, pm_state):
        # Make action selection impossible if the PM would exceed capacity
        step = self.current_step if self.current_step < self.step_limit else self.step_limit-1
        data_center = np.vstack([pm_state, self.demand[step]])
        data_center = np.where(data_center>1,1,data_center) # Fix rounding errors
        self.state["state"] = data_center
        self.state["action_mask"] = np.ones(self.n_pms)
        self.state["avail_actions"] = np.ones(self.n_pms)
        if self.mask:
            action_mask = (pm_state[:, 1:] + self.demand[step, 1:]) <= 1
            self.state["action_mask"] = (action_mask.sum(axis=1)==2).astype(int)
        
    def _RESET(self):
        self.current_step = 0
        self.assignment = {}
        self.demand = self.generate_demand()
        self.durations = generate_durations(self.demand)
        self.state = (np.zeros((self.n_pms, 3)), self.demand[0])
        return self.state

    def step(self, action):
        return self._STEP(action)

    def reset(self):
        return self._RESET()

def generate_durations(demand):
    # duration_params = np.array([ 6.53563303e-02,  5.16222242e+01,  4.05028032e+06, -4.04960880e+06])
    return {i: np.random.randint(low=i+1, high=len(demand)+1)
        for i, j in enumerate(demand)}

def gaussian_model(params, x):
    return params[2] * np.exp(-0.5*((x - params[0]) / params[1]) ** 2) + params[3]


import numpy as np
import gym
from gym import spaces, logger
from gym.utils import seeding
from or_gym.utils import assign_env_config
from copy import copy

class PortfolioOptEnv(gym.Env): 
    '''
    Portfolio Optimization Problem 

    Instance: Multi-Period Asset Allocation Problem, Dantzing & Infager, 1993 

    The Portfolio Optimization (PO) Problem is a problem that seeks to optimize 
    the distribution of assets in a financial portfolio to with respect to a desired 
    financial metric (e.g. maximal return, minimum risk, etc.). 

    In this particular instance by Dantzing & Infager, the optimizer begins with a 
    quantity of cash and has the opportunity to purchase or sell 3 other assets in each  
    of 10 different investement periods. Each transaction incurs a cost and prices of 
    the 3 assets are subject to change over time. Cash value is consant (price = 1). 
    The objective is to maximize the amount of wealth (i.e. the sum total of asset values)
    at the end of the total investment horizon.

    The episodes proceed by the optimizer deciding whether to buy or sell each asset
    in each time period. The episode ends when either all 10 periods have passed or 
    if the amount of any given asset held becomes negative.  

    Observation:
        Type: Box(9)
        "asset prices" (idx 0, 1, 2, 3): array of asset prices [cash, asset1, asset2, asset3]
        "asset quantities" (idx 4, 5, 6, 7): array of asset quantities [cash, asset1, asset2, asset3]
        "total wealth" (idx 8): current total wealth (sum of price*quantity for each asset)
    

    Actions:
        Type: Box (3)
        "asset 1 transaction amount" (idx 0): x in [-2000, 2000]: Buy (positive) or sell (negative) x shares of asset 1; 
        "asset 2 transaction amount" (idx 1): x in [-2000, 2000]: Buy (positive) or sell (negative) x shares of asset 2; 
        "asset 3 transaction amount" (idx 2): x in [-2000, 2000]: Buy (positive) or sell (negative) x shares of asset 3; 

    Reward:
        Change in total wealth from previous period or [-max(asset price of all assets) *  maximum transaction size]
        if an asset quantity becomes negative, at which 
        point the episode ends.

    Starting State:
        Starting amount of cash and wealth and prices. 

    Episode Termination:
        Negative asset quantity or traversal of investment horizon. 
    '''    
    def __init__(self, *args, **kwargs):
        self.num_assets = 3 # Number of assets 
        self.initial_cash = 100 # Starting amount of capital 
        self.step_limit = 10 # Investment horizon 

        self.cash = copy(self.initial_cash)

        #Transaction costs proportional to amount bought 
        self.buy_cost = np.array([0.045, 0.025, 0.035])
        self.sell_cost = np.array([0.04, 0.02, 0.03])
        # self.step_limit = 10
        # assign_env_config(self, kwargs)
        # self.asset_price_means = asset_price_means.T
        # # self.asset_price_means = (np.random.randint(10, 50, self.num_assets) \
        # #     * np.ones((self.step_limit, self.num_assets))).T
        # self.asset_price_var = np.ones(self.asset_price_means.shape)

        # Prices of assets have a mean value in every period and vary according to a Gaussian distribution 
        asset1mean = np.array([1.25, 2, 4, 5, 3, 2, 3, 6, 9, 7]).reshape(1, -1)  # Up and down all the way 
        asset2mean = np.array([5, 3, 2, 2, 1.25, 4, 5, 6, 7, 8]).reshape(1, -1)  # Down intially then up 
        asset3mean = np.array([3, 5, 6, 9, 10, 8, 4, 2, 1.25, 4]).reshape(1, -1) # Up initially then down 
        self.asset_price_means = np.vstack([asset1mean, asset2mean, asset3mean])
        self.asset_price_var = np.ones((self.asset_price_means.shape)) * 0.45
        
        # Cash on hand, asset prices, num of shares, portfolio value
        self.obs_length = 1 + 2 * self.num_assets

        self.observation_space = spaces.Box(-20000, 20000, shape=(self.obs_length,))
        self.action_space = spaces.Box(-2000, 2000, shape=(self.num_assets,))
        
        self.reset()
        
    def _RESET(self):
        self.step_count = 0
        self.asset_prices = self._generate_asset_prices()
        self.holdings = np.zeros(self.num_assets)
        self.cash = copy(self.initial_cash)
        self.state = np.hstack([
            self.initial_cash,
            self.asset_prices[:, self.step_count],
            self.holdings])
        return self.state
    
    def _generate_asset_prices(self):
        asset_prices = np.array([np.random.normal(mu, sig) for mu, sig in 
            zip(self.asset_price_means.flatten(), self.asset_price_var.flatten())]
            ).reshape(self.asset_price_means.shape)
        # Zero out negative asset prices and all following prices - implies
        # equity is bankrupt and worthless.
        zero_vals = np.vstack(np.where(asset_prices<0))
        cols = np.unique(zero_vals[0])
        for c in cols:
            first_zero = zero_vals[1][np.where(zero_vals[0]==c)[0].min()]
            asset_prices[c,first_zero:] = 0
        return asset_prices
    
    def _STEP(self, action):
        
        assert self.action_space.contains(action)
    
        asset_prices = self.asset_prices[:, self.step_count].copy()
        
        for idx, a in enumerate(action):
            if a == 0:
                continue
            # Sell a shares of asset
            elif a < 0:
                a = np.abs(a)
                if a > self.holdings[idx]:
                    a = self.holdings[idx]
                self.holdings[idx] -= a
                self.cash += asset_prices[idx] * a * (1 - self.sell_cost[idx])
            # Buy a shares of asset
            elif a > 0:
                purchase_cost = asset_prices[idx] * a * (1 + self.buy_cost[idx])
                if self.cash < purchase_cost:
                    a = np.floor(self.cash / (
                        asset_prices[idx] * (1 + self.buy_cost[idx])))
                    purchase_cost = asset_prices[idx] * a * (1 + self.buy_cost[idx])
                self.holdings[idx] += a
                self.cash -= purchase_cost
                
        # Return total portfolio value at the end of the horizon as reward
        if self.step_count + 1 == self.step_limit: 
            reward = np.dot(asset_prices, self.holdings) + self.cash
        else: 
            reward = 0 
        self.step_count += 1

        # Finish if 10 periods have passed - end of investment horizon 
        if self.step_count >= self.step_limit:
            done = True
        else:
            self._update_state()
            done = False
            
        return self.state, reward, done, {}
    
    def _update_state(self):
        self.state = np.hstack([
            self.cash,
            self.asset_prices[:, self.step_count],
            self.holdings
        ])

    def step(self, action):
        return self._STEP(action)

    def reset(self):
        return self._RESET()



from gym.envs.registration import register

# Knapsack Environments
register(id='Knapsack-v0',
    entry_point='or_gym.envs.classic_or.knapsack:KnapsackEnv'
)

register(id='Knapsack-v1',
	entry_point='or_gym.envs.classic_or.knapsack:BinaryKnapsackEnv'
)

register(id='Knapsack-v2',
	entry_point='or_gym.envs.classic_or.knapsack:BoundedKnapsackEnv'
)

register(id='Knapsack-v3',
	entry_point='or_gym.envs.classic_or.knapsack:OnlineKnapsackEnv'
)

# Bin Packing Environments
register(id='BinPacking-v0',
	entry_point='or_gym.envs.classic_or.binpacking:BinPackingEnv'
)

register(id='BinPacking-v1',
	entry_point='or_gym.envs.classic_or.binpacking:BinPackingLW1'
)

register(id='BinPacking-v2',
	entry_point='or_gym.envs.classic_or.binpacking:BinPackingPP0'
)

register(id='BinPacking-v3',
	entry_point='or_gym.envs.classic_or.binpacking:BinPackingPP1'
)

register(id='BinPacking-v4',
	entry_point='or_gym.envs.classic_or.binpacking:BinPackingBW0'
)

register(id='BinPacking-v5',
	entry_point='or_gym.envs.classic_or.binpacking:BinPackingBW1'
)

# Newsvendor Envs
register(id='Newsvendor-v0',
	entry_point='or_gym.envs.classic_or.newsvendor:NewsvendorEnv'
)

# Virtual Machine Packing Envs
register(id='VMPacking-v0',
	entry_point='or_gym.envs.classic_or.vmpacking:VMPackingEnv'
)

register(id='VMPacking-v1',
	entry_point='or_gym.envs.classic_or.vmpacking:TempVMPackingEnv'
)

# Vehicle Routing Envs
register(id='VehicleRouting-v0',
	entry_point='or_gym.envs.classic_or.vehicle_routing:VehicleRoutingEnv'
)

# TSP
register(id='TSP-v0',
	entry_point='or_gym.envs.classic_or.tsp:TSPEnv'
)

register(id='TSP-v1',
	entry_point='or_gym.envs.classic_or.tsp:TSPDistCost'
)

# Inventory Management Envs
register(id='InvManagement-v0',
	entry_point='or_gym.envs.supply_chain.inventory_management:InvManagementBacklogEnv'
)

register(id='InvManagement-v1',
	entry_point='or_gym.envs.supply_chain.inventory_management:InvManagementLostSalesEnv'
)

register(id='NetworkManagement-v0',
	entry_point='or_gym.envs.supply_chain.network_management:NetInvMgmtBacklogEnv'
)

register(id='NetworkManagement-v1',
	entry_point='or_gym.envs.supply_chain.network_management:NetInvMgmtLostSalesEnv'
)

# Asset Allocation Envs
register(id='PortfolioOpt-v0',
	entry_point='or_gym.envs.finance.portfolio_opt:PortfolioOptEnv'
)


#!usr/bin/env python

from setuptools import setup, find_packages
import sys
import os

sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'or_gym'))
from or_gym.version import VERSION

setup(name='or-gym',
	version=VERSION,
	description='OR-Gym: A set of environments for developing reinforcement learning agents for OR problems.',
	author='Christian Hubbs, Hector Perez Parra, Owais Sarwar',
	license='MIT',
	url='https://github.com/hubbs5/or-gym',
	packages=find_packages(),
	install_requires=[
		'gym>=0.15.0',
		'numpy>=1.16.1',
		'scipy>=1.0',
		'matplotlib>=3.1',
		'networkx>=2.3'],
	zip_safe=False,
	python_requires='>=3.5',
	classifiers=[
		'Development Status :: 3 - Alpha',
		'Intended Audience :: Developers',
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.5',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
		'Programming Language :: Python :: 3.8',
	]
)



VERSION='0.2.0'

ENV_LIST = ['Knapsack-v0', 'Knapsack-v1', 'Knapsack-v2',
            'BinPacking-v0', 'BinPacking-v1', 'BinPacking-v2',
            'BinPacking-v3', 'BinPacking-v4', 'BinPacking-v5',
            'VMPacking-v0',
            'InvManagement-v0', 'InvManagement-v1',
            'NetworkManagement-v0', 'NetworkManagement-v1',
            'PortfolioOpt-v0',
            'VehicleRouting-v0', 
            'Newsvendor-v0',
            'TSP-v0', 'TSP-v1']


import numpy as np

def assign_env_config(self, kwargs):
    for key, value in kwargs.items():
        setattr(self, key, value)
    if hasattr(self, 'env_config'):
        # print(self.env_config)
        for key, value in self.env_config.items():
            # Check types based on default settings
            if hasattr(self, key):
                if type(getattr(self,key)) == np.ndarray:
                    setattr(self, key, value)
                else:
                    setattr(self, key,
                        type(getattr(self, key))(value))
            else:
                setattr(self, key, value)
                
# Get Ray to work with gym registry
def create_env(config, *args, **kwargs):
	if type(config) == dict:
		env_name = config['env']
	else:
		env_name = config
	if env_name == 'Knapsack-v0':
		from or_gym.envs.classic_or.knapsack import KnapsackEnv as env
	elif env_name == 'Knapsack-v1':
		from or_gym.envs.classic_or.knapsack import BinaryKnapsackEnv as env
	elif env_name == 'Knapsack-v2':
		from or_gym.envs.classic_or.knapsack import BoundedKnapsackEnv as env
	elif env_name == 'Knapsack-v3':
		from or_gym.envs.classic_or.knapsack import OnlineKnapsackEnv as env
	elif env_name == 'BinPacking-v0':
		from or_gym.envs.classic_or.binpacking import BinPackingEnv as env
	elif env_name == 'BinPacking-v1':
		from or_gym.envs.classic_or.binpacking import BinPackingLW1 as env
	elif env_name == 'BinPacking-v2':
		from or_gym.envs.classic_or.binpacking import BinPackingPP0 as env
	elif env_name == 'BinPacking-v3':
		from or_gym.envs.classic_or.binpacking import BinPackingPP1 as env
	elif env_name == 'BinPacking-v4':
		from or_gym.envs.classic_or.binpacking import BinPackingBW0 as env
	elif env_name == 'BinPacking-v5':
		from or_gym.envs.classic_or.binpacking import BinPackingBW1 as env
	elif env_name == 'VMPacking-v0':
		from or_gym.envs.classic_or.vmpacking import VMPackingEnv as env
	elif env_name == 'VMPacking-v1':
		from or_gym.envs.classic_or.vmpacking import TempVMPackingEnv as env
	elif env_name == 'PortfolioOpt-v0':
		from or_gym.envs.finance.portfolio_opt import PortfolioOptEnv as env
	elif env_name == 'TSP-v0':
		raise NotImplementedError('{} not yet implemented.'.format(env_name))
	elif env_name == 'VehicleRouting-v0':
		from or_gym.envs.classic_or.vehicle_routing import VehicleRoutingEnv as env
	elif env_name == 'VehicleRouting-v1':
		from or_gym.envs.classic_or.vehicle_routing import VehicleRoutingEnv as env
	elif env_name == 'NewsVendor-v0':
		from or_gym.envs.classic_or.newsvendor import NewsvendorEnv as env
	elif env_name == 'InvManagement-v0':
		from or_gym.envs.supply_chain.inventory_management import InvManagementBacklogEnv as env
	elif env_name == 'InvManagement-v1':
		from or_gym.envs.supply_chain.inventory_management import InvManagementLostSalesEnv as env
	else:
		raise NotImplementedError('Environment {} not recognized.'.format(env_name))
	return env



import os
import sys
import warnings

from gym import error
from or_gym.version import VERSION as __version__
from or_gym.utils import *

from gym.core import Env, GoalEnv, Wrapper, ObservationWrapper, ActionWrapper, RewardWrapper
from gym.envs import make, spec, register
from or_gym.envs import classic_or